# Copyright 2011-2015 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import unicode_literals
import datetime

import isodate
import six


# Some type conversion functions
def to_date(value):
    return isodate.parse_date(value)


def to_time(value):
    return isodate.parse_time(value)


def to_datetime(value):
    # We encountered situations where the T separator was replaces by a space
    # so make sure that is not the case
    value = value.replace(' ', 'T')
    return isodate.parse_datetime(value)


def to_string(value):
    """
    For Python 2, make sure the string is properly converted to unicode

    :param basestring value:
    :return:
    """

    if six.PY2:
        if isinstance(value, unicode):
            value.encode('utf8')
        elif isinstance(value, str):
            # Must be encoded in UTF-8
            value = value.decode('utf8')
    else:
        value = str(value)

    return value


def to_timedelta(value):
    return isodate.parse_duration(value).tdelta


def to_bool(value):
    return value in ["true", "1"]


def from_datetime(value):
    if isinstance(value, six.string_types):
        to_datetime(value)  # First make sure it is a datetime

    if isinstance(value, datetime.datetime):
        return value.isoformat()
    else:
        raise ValueError('To create a proper string representation for a'
                         ' datetime, either a datetime.datetime or str has'
                         ' to be supplied!')


def from_date(value):
    if isinstance(value, six.string_types):
        value = isodate.parse_date(value)

    if isinstance(value, datetime.date):
        return value.isoformat()
    else:
        raise ValueError('To create a proper string representation for a date,'
                         ' either a datetime.date or str has to be supplied!')


def from_time(value):
    if isinstance(value, six.string_types):
        value = isodate.parse_time(value)

    if isinstance(value, datetime.time):
        return value.isoformat()
    else:
        raise ValueError('To create a proper string representation for a time,'
                         ' either a datetime.time or str has to be supplied!')


def from_timedelta(value):
    if isinstance(value, six.string_types):
        value = isodate.parse_duration(value)
    elif isinstance(value, datetime.timedelta):
        value = isodate.duration.Duration(days=value.days,
                                          seconds=value.seconds,
                                          microseconds=value.microseconds)

    if isinstance(value, isodate.duration.Duration):
        return isodate.duration_isoformat(value)
    else:
        raise ValueError('To create a proper string representation for a duration,'
                         ' either a isodate.duration.Duration or str has to be supplied!')


def from_bool(value):
    if isinstance(value, six.string_types):
        if value in ["true", "false", "1", "0"]:
            return value
        else:
            raise ValueError('Value {} is not a valid string representation of a bool'.format(value))
    elif isinstance(value, bool):
        return 'true' if value else 'false'
    else:
        raise TypeError('To create a proper string presentation for a bool,'
                        ' either a bool or str has to be supplied!')


def from_int(value):
    if not isinstance(value, int):
        value = int(value)
    return six.text_type(value)


def from_float(value):
    if not isinstance(value, float):
        value = float(value)
    return six.text_type(value)


def from_string(value):
    """
    For Python 2, make sure the string is a valid utf-8 encoded str before
    shipping it off to urllib and such.

    :param basestring value:
    :return:
    """
    if not isinstance(value, six.string_types):
        value = str(value)

    if six.PY2:
        if isinstance(value, unicode):
            value.encode('utf8')
        elif isinstance(value, str):
            # Must be encoded in UTF-8
            value = value.decode('utf8')
    else:
        value = str(value)

    return value


# Here to be after all needed function definitions
TYPE_TO_MAP = {
    'xs:anyURI': to_string,
    'xs:string': to_string,
    'xs:boolean': to_bool,
    'xs:integer': int,
    'xs:long': int,
    'xs:float': float,
    'xs:double': float,
    'xs:dateTime': to_datetime,
    'xs:time': to_time,
    'xs:date': to_date,
    'xs:duration': to_timedelta,
}

TYPE_FROM_MAP = {
    'xs:anyURI': from_string,
    'xs:string': from_string,
    'xs:boolean': from_bool,
    'xs:integer': from_int,
    'xs:long': from_int,
    'xs:float': from_float,
    'xs:double': from_float,
    'xs:dateTime': from_datetime,
    'xs:time': from_time,
    'xs:date': from_date,
    'xs:duration': from_timedelta,
}

TYPE_TO_PYTHON = {
    'xs:anyURI': str,
    'xs:string': str,
    'xs:boolean': bool,
    'xs:integer': int,
    'xs:long': int,
    'xs:float': float,
    'xs:double': float,
    'xs:dateTime': datetime.datetime,
    'xs:time': datetime.time,
    'xs:date': datetime.date,
    'xs:duration': datetime.timedelta,
}


def convert_to(value, type_):
    return TYPE_TO_MAP.get(type_, six.text_type)(value)


def convert_from(value, type_):
    return TYPE_FROM_MAP[type_](value)



# Copyright 2011-2015 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import unicode_literals
import datetime
import io
import netrc
import os
import re
import threading

from progressbar import AdaptiveETA, AdaptiveTransferSpeed, Bar, BouncingBar, \
    DataSize, Percentage, ProgressBar, Timer, UnknownLength
import requests
import six
from six.moves.urllib import parse

from . import exceptions
from .core import XNATListing, caching
from .inspect import Inspect
from .prearchive import Prearchive
from .users import Users
from .services import Services
from .exceptions import XNATValueError, XNATNotConnectedError

try:
    FILE_TYPES = (file, io.IOBase)
except NameError:
    FILE_TYPES = io.IOBase


class XNATSession(object):
    """
    The main XNATSession session class. It keeps a connection to XNATSession alive and
    manages the main communication to XNATSession. To keep the connection alive
    there is a background thread that sends a heart-beat to avoid a time-out.

    The main starting points for working with the XNATSession server are:

    * :py:meth:`XNATSession.projects <xnat.session.XNATSession.projects>`
    * :py:meth:`XNATSession.subjects <xnat.session.XNATSession.subjects>`
    * :py:meth:`XNATSession.experiments <xnat.session.XNATSession.experiments>`
    * :py:meth:`XNATSession.prearchive <xnat.session.XNATSession.prearchive>`
    * :py:meth:`XNATSession.services <xnat.session.XNATSession.services>`
    * :py:meth:`XNATSession.users <xnat.session.XNATSession.users>`

    .. note:: Some methods create listing that are using the :py:class:`xnat.core.XNATListing <xnat.core.XNATListing>`
              class. They allow for indexing with both XNATSession ID and a secondary key (often the
              label). Also they support basic filtering and tabulation.

    There are also methods for more low level communication. The main methods
    are :py:meth:`XNATSession.get <xnat.session.XNATSession.get>`, :py:meth:`XNATSession.post <xnat.session.XNATSession.post>`,
    :py:meth:`XNATSession.put <xnat.session.XNATSession.put>`, and :py:meth:`XNATSession.delete <xnat.session.XNATSession.delete>`.
    The methods do not query URIs but instead query XNATSession REST paths as described in the
    `XNATSession 1.6 REST API Directory <https://wiki.xnat.org/display/XNAT16/XNATSession+REST+API+Directory>`_.

    For an even lower level interfaces, the :py:attr:`XNATSession.interface <xnat.session.XNATSession.interface>`
    gives access to the underlying `requests <https://requests.readthedocs.org>`_ interface.
    This interface has the user credentials and benefits from the keep alive of this class.

    .. note:: :py:class:`XNATSession <xnat.session.XNATSession>` Objects have a client-side cache. This is for efficiency, but might cause
              problems if the server is being changed by a different client. It is possible
              to clear the current cache using :py:meth:`XNATSession.clearcache <xnat.session.XNATSession.clearcache>`.
              Turning off caching complete can be done by setting
              :py:attr:`XNATSession.caching <xnat.session.XNATSession.caching>`.

    .. warning:: You should NOT try use this class directly, it should only
                 be created by :py:func:`xnat.connect <xnat.connect>`.
    """

    def __init__(self, server, logger, interface=None, user=None,
                 password=None, keepalive=None, debug=False,
                 original_uri=None, logged_in_user=None, default_timeout=300):
        # Class lookup to populate (session specific, as all session have their
        # own classes based on the server xsd)
        self.XNAT_CLASS_LOOKUP = {}

        self.classes = None
        self._interface = interface
        self._projects = None
        self._server = parse.urlparse(server) if server else None
        if original_uri is not None:
            self._original_uri = original_uri.rstrip('/')
        else:
            self._original_uri = server.rstrip('/')
        self._logged_in_user = logged_in_user
        self._cache = {'__objects__': {}}
        self.caching = True
        self._source_code_file = None
        self._services = Services(xnat_session=self)
        self._prearchive = Prearchive(xnat_session=self)
        self._users = Users(xnat_session=self)
        self._debug = debug
        self.logger = logger
        self.inspect = Inspect(self)
        self.request_timeout = default_timeout

        # Accepted status
        self.accepted_status_get = [200]
        self.accepted_status_post = [200, 201]
        self.accepted_status_put = [200, 201]
        self.accepted_status_delete = [200]
        self.skip_response_check = False
        self.skip_response_content_check = False

        session_expiration = self.session_expiration_time
        if session_expiration is not None:
            # 30 seconds before the expiration, at most once per 10 seconds
            if session_expiration[1] < 30:
                self.logger.warning(
                    ('Server session expiration time ({}) is lower than 30 seconds,'
                     ' setting heartbeat interval to the minimum of 10 seconds.').format(session_expiration[1]))
            default_keepalive = max(session_expiration[1] - 20, 10)
        else:
            default_keepalive = 14 * 60  # Default to 14 minutes

        # Set the keep alive settings and spawn the keepalive thread for sending heartbeats
        if keepalive is None or keepalive is True:
            keepalive = default_keepalive

        if isinstance(keepalive, int) and keepalive > 0:
            self._keepalive = True
            self._keepalive_interval = keepalive
        else:
            self._keepalive = False
            self._keepalive_interval = default_keepalive  # Not used while keepalive is false, but set a default

        self._keepalive_running = False
        self._keepalive_thread = None
        self._keepalive_event = threading.Event()

        # If needed connect here
        self.connect(server=server, user=user, password=password)

    def __del__(self):
        self.disconnect()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.disconnect()

    def connect(self, server=None, user=None, password=None):
        # If not connected, connect now
        if self.interface is None:
            if server is None:
                raise ValueError('Cannot connect if no server is given')
            self.logger.info('Connecting to server {}'.format(server))
            if self._interface is not None:
                self.disconnect()

            self._server = parse.urlparse(server)

            if user is None and password is None:
                self.logger.info('Retrieving login info for {}'.format(self._server.netloc))
                try:
                    user, _, password = netrc.netrc().authenticators(self._server.netloc)
                except TypeError:
                    raise ValueError('Could not retrieve login info for "{}" from the .netrc file!'.format(server))

            self._interface = requests.Session()
            if (user is not None) or (password is not None):
                self._interface.auth = (user, password)

        # Create a keepalive thread
        self._keepalive_running = True
        self._keepalive_thread = threading.Thread(target=self._keepalive_thread_run, name='XNATpyKeepAliveThread')
        self._keepalive_thread.daemon = True  # Make sure thread stops if program stops
        self._keepalive_thread.start()
        self.heartbeat()  # Make sure the heartbeat is given and there is no chance of timeout

    def disconnect(self):
        # Stop the keepalive thread
        self._keepalive_running = False
        self._keepalive_event.set()

        if self._keepalive_thread is not None:
            if self._keepalive_thread.is_alive():
                self._keepalive_thread.join(3.0)
            self._keepalive_thread = None

        # Kill the session
        if self._server is not None and self._interface is not None:
            self.delete('/data/JSESSION', headers={'Connection': 'close'})

        # Set the server and interface to None
        self._interface = None
        self._server = None

        # If this object is created using an automatically generated file
        # we have to remove it.
        if self._source_code_file is not None:
            source_pyc = self._source_code_file + 'c'
            if os.path.isfile(self._source_code_file):
                os.remove(self._source_code_file)
                self._source_code_file = None
            if os.path.isfile(source_pyc):
                os.remove(source_pyc)

        self.classes = None

    @property
    def keepalive(self):
        return self._keepalive

    @keepalive.setter
    def keepalive(self, value):
        if isinstance(value, int):
            if value > 0:
                self._keepalive_interval = value
                value = True
            else:
                value = False

        elif not isinstance(value, bool):
            raise TypeError('Type should be an integer or boolean!')

        self._keepalive = value

        if self.keepalive:
            # Send a new heartbeat and restart the timer to make sure the interval is correct
            self._keepalive_event.set()
            self.heartbeat()

    def heartbeat(self):
        self.get('/data/JSESSION')

    def _keepalive_thread_run(self):
        # This thread runs until the program stops, it should be inexpensive if not used due to the long sleep time
        while self._keepalive_running:
            # Wait returns False on timeout and True otherwise
            if not self._keepalive_event.wait(self._keepalive_interval):
                if self.keepalive:
                    self.heartbeat()
            else:
                self._keepalive_event.clear()

    @property
    def logged_in_user(self):
        return self._logged_in_user

    @property
    def debug(self):
        return self._debug

    @property
    def interface(self):
        """
        The underlying `requests <https://requests.readthedocs.org>`_ interface used.
        """
        return self._interface

    @property
    def uri(self):
        return '/data/archive'

    @property
    def fulluri(self):
        return self.uri

    @property
    def xnat_session(self):
        return self

    @property
    def session_expiration_time(self):
        """
        Get the session expiration time information from the cookies. This
        returns the timestamp (datetime format) when the session was created
        and an integer with the session timeout interval.

        This can return None if the cookie is not found or cannot be parsed.

        :return: datetime with last session refresh and integer with timeout in seconds
        :rtype: tuple
        """
        expiration_string = self.interface.cookies.get('SESSION_EXPIRATION_TIME')

        if expiration_string is None:
            return

        match = re.match(r'^"(?P<timestamp>\d+),(?P<interval>\d+)"$', expiration_string)
        if match is None:
            self.logger.warning('Could not parse SESSION_EXPIRATION_TIME cookie')
            return None

        session_timestamp = datetime.datetime.fromtimestamp(int(match.group('timestamp')) / 1000)
        expiration_interval = int(match.group('interval')) / 1000
        return session_timestamp, expiration_interval

    def _check_response(self, response, accepted_status=None, uri=None):
        if self.debug:
            self.logger.debug('Received response with status code: {}'.format(response.status_code))

        if not self.skip_response_check:
            if accepted_status is None:
                accepted_status = [200, 201, 202, 203, 204, 205, 206]  # All successful responses of HTML
            if response.status_code not in accepted_status or (not self.skip_response_content_check and response.text.startswith(('<!DOCTYPE', '<html>'))):
                raise exceptions.XNATResponseError('Invalid response from XNATSession for url {} (status {}):\n{}'.format(uri, response.status_code, response.text))

    def _check_connection(self):
        """
        Check if connection is still open
        """
        if self.interface is None:
            message = 'Not connected to server. Either the connection was not established or was closed!'
            self.logger.error(message)
            raise XNATNotConnectedError(message)

    def get(self, path, format=None, query=None, accepted_status=None, timeout=None, headers=None):
        """
        Retrieve the content of a given REST directory.

        :param str path: the path of the uri to retrieve (e.g. "/data/archive/projects")
                         the remained for the uri is constructed automatically
        :param str format: the format of the request, this will add the format= to the query string
        :param dict query: the values to be added to the query string in the uri
        :param list accepted_status: a list of the valid values for the return code, default [200]
        :param timeout: timeout in seconds, float or (connection timeout, read timeout)
        :type timeout: float or tuple
        :param dict headers: the HTTP headers to include
        :returns: the requests reponse
        :rtype: requests.Response
        """
        self._check_connection()

        accepted_status = accepted_status or self.accepted_status_get
        uri = self._format_uri(path, format, query=query)
        timeout = timeout or self.request_timeout

        self.logger.debug('GET URI {}'.format(uri))

        try:
            response = self.interface.get(uri, timeout=timeout, headers=headers)
        except requests.exceptions.SSLError:
            raise exceptions.XNATSSLError('Encountered a problem with the SSL connection, are you sure the server is offering https?')
        self._check_response(response, accepted_status=accepted_status, uri=uri)  # Allow OK, as we want to get data
        return response

    def head(self, path, accepted_status=None, allow_redirects=False, timeout=None, headers=None):
        """
        Retrieve the header for a http request of a given REST directory.

        :param str path: the path of the uri to retrieve (e.g. "/data/archive/projects")
                         the remained for the uri is constructed automatically
        :param list accepted_status: a list of the valid values for the return code, default [200]
        :param bool allow_redirects: allow you request to be redirected
        :param timeout: timeout in seconds, float or (connection timeout, read timeout)
        :type timeout: float or tuple
        :param dict headers: the HTTP headers to include
        :returns: the requests reponse
        :rtype: requests.Response
        """
        self._check_connection()

        accepted_status = accepted_status or self.accepted_status_get
        uri = self._format_uri(path)
        timeout = timeout or self.request_timeout

        self.logger.debug('GET URI {}'.format(uri))

        try:
            response = self.interface.head(uri, allow_redirects=allow_redirects, timeout=timeout, headers=headers)
        except requests.exceptions.SSLError:
            raise exceptions.XNATSSLError('Encountered a problem with the SSL connection, are you sure the server is offering https?')
        self._check_response(response, accepted_status=accepted_status, uri=uri)  # Allow OK, as we want to get data
        return response

    def post(self, path, data=None, json=None, format=None, query=None, accepted_status=None, timeout=None, headers=None):
        """
        Post data to a given REST directory.

        :param str path: the path of the uri to retrieve (e.g. "/data/archive/projects")
                         the remained for the uri is constructed automatically
        :param data: Dictionary, bytes, or file-like object to send in the body of the :class:`Request`.
        :param json: json data to send in the body of the :class:`Request`.
        :param str format: the format of the request, this will add the format= to the query string
        :param dict query: the values to be added to the query string in the uri
        :param list accepted_status: a list of the valid values for the return code, default [200, 201]
        :param timeout: timeout in seconds, float or (connection timeout, read timeout)
        :type timeout: float or tuple
        :param dict headers: the HTTP headers to include
        :returns: the requests reponse
        :rtype: requests.Response
        """
        self._check_connection()

        accepted_status = accepted_status or self.accepted_status_post
        uri = self._format_uri(path, format, query=query)
        timeout = timeout or self.request_timeout

        self.logger.debug('POST URI {}'.format(uri))
        if self.debug:
            self.logger.debug('POST DATA {}'.format(data))

        try:
            response = self._interface.post(uri, data=data, json=json, timeout=timeout, headers=headers)
        except requests.exceptions.SSLError:
            raise exceptions.XNATSSLError('Encountered a problem with the SSL connection, are you sure the server is offering https?')
        self._check_response(response, accepted_status=accepted_status, uri=uri)
        return response

    def put(self, path, data=None, files=None, json=None, format=None, query=None, accepted_status=None, timeout=None, headers=None):
        """
        Put the content of a given REST directory.

        :param str path: the path of the uri to retrieve (e.g. "/data/archive/projects")
                         the remained for the uri is constructed automatically
        :param data: Dictionary, bytes, or file-like object to send in the body of the :class:`Request`.
        :param json: json data to send in the body of the :class:`Request`.
        :param files: Dictionary of ``'name': file-like-objects`` (or ``{'name': file-tuple}``) for multipart encoding upload.
                      ``file-tuple`` can be a 2-tuple ``('filename', fileobj)``, 3-tuple ``('filename', fileobj, 'content_type')``
                      or a 4-tuple ``('filename', fileobj, 'content_type', custom_headers)``, where ``'content-type'`` is a string
                      defining the content type of the given file and ``custom_headers`` a dict-like object containing additional headers
                      to add for the file.
        :param str format: the format of the request, this will add the format= to the query string
        :param dict query: the values to be added to the query string in the uri
        :param list accepted_status: a list of the valid values for the return code, default [200, 201]
        :param timeout: timeout in seconds, float or (connection timeout, read timeout)
        :type timeout: float or tuple
        :param dict headers: the HTTP headers to include
        :returns: the requests reponse
        :rtype: requests.Response
        """
        self._check_connection()

        accepted_status = accepted_status or self.accepted_status_put
        uri = self._format_uri(path, format, query=query)
        timeout = timeout or self.request_timeout

        self.logger.debug('PUT URI {}'.format(uri))
        if self.debug:
            self.logger.debug('PUT DATA {}'.format(data))
            self.logger.debug('PUT FILES {}'.format(data))

        try:
            response = self._interface.put(uri, data=data, files=files, json=json, timeout=timeout, headers=headers)
        except requests.exceptions.SSLError:
            raise exceptions.XNATSSLError('Encountered a problem with the SSL connection, are you sure the server is offering https?')
        self._check_response(response, accepted_status=accepted_status, uri=uri)  # Allow created OK or Create status (OK if already exists)
        return response

    def delete(self, path, headers=None, accepted_status=None, query=None, timeout=None):
        """
        Delete the content of a given REST directory.

        :param str path: the path of the uri to retrieve (e.g. "/data/archive/projects")
                         the remained for the uri is constructed automatically
        :param dict headers: the HTTP headers to include
        :param dict query: the values to be added to the query string in the uri
        :param list accepted_status: a list of the valid values for the return code, default [200]
        :param timeout: timeout in seconds, float or (connection timeout, read timeout)
        :type timeout: float or tuple
        :returns: the requests reponse
        :rtype: requests.Response
        """
        self._check_connection()

        accepted_status = accepted_status or self.accepted_status_delete
        uri = self._format_uri(path, query=query)
        timeout = timeout or self.request_timeout

        self.logger.debug('DELETE URI {}'.format(uri))
        if self.debug:
            self.logger.debug('DELETE HEADERS {}'.format(headers))

        try:
            response = self.interface.delete(uri, headers=headers, timeout=timeout)
        except requests.exceptions.SSLError:
            raise exceptions.XNATSSLError('Encountered a problem with the SSL connection, are you sure the server is offering https?')
        self._check_response(response, accepted_status=accepted_status, uri=uri)
        return response

    def _format_uri(self, path, format=None, query=None, scheme=None):
        if path[0] != '/':

            if self._original_uri is not None and path.startswith(self._original_uri):
                path = path[len(self._original_uri):]  # Strip original uri

            if path[0] != '/':
                raise XNATValueError('The requested URI path should start with a / (e.g. /data/projects), found {}'.format(path))

        if query is None:
            query = {}

        if format is not None:
            query['format'] = format

        # Sanitize unicode in query
        if six.PY2:
            query = {k: v.encode('utf-8', 'xmlcharrefreplace') if isinstance(v, unicode) else v for k, v in query.items()}

        # Create the query string
        if len(query) > 0:
            query_string = parse.urlencode(query, doseq=True)
        else:
            query_string = ''

        data = (scheme or self._server.scheme,
                self._server.netloc,
                self._server.path.rstrip('/') + path,
                '',
                query_string,
                '')

        return parse.urlunparse(data)

    def url_for(self, obj, query=None, scheme=None):
        """
        Return the (external) url for a given XNAT object
        :param XNATBaseObject obj: object to get url for
        :param query: extra query string parameters
        :param scheme: scheme to use (when not using original url scheme)
        :return: external url for the object
        """
        return self._format_uri(obj.fulluri, query=query, scheme=scheme)

    def get_json(self, uri, query=None, accepted_status=None):
        """
        Helper function that perform a GET, but sets the format to JSON and
        parses the result as JSON

        :param str uri: the path of the uri to retrieve (e.g. "/data/archive/projects")
                         the remained for the uri is constructed automatically
        :param dict query: the values to be added to the query string in the uri
        """
        response = self.get(uri, format='json', query=query, accepted_status=accepted_status)
        try:
            return response.json()
        except ValueError:
            # Multiple options to support newer XNAT versions
            if response.text.startswith((
                '<?xml version="1.0" encoding="UTF-8"?>\n<cat:Catalog',
                '<?xml version="1.0" encoding="UTF-8"?>\n<cat:DCMCatalog',  # XNAT 1.7.5.3+
            )):
                # Probably XML catalog for resource
                parts = uri.rsplit('/resources/')
                if len(parts) != 2:
                    raise XNATValueError('Could not decode JSON from [{}] and could not figure out resource URI'.format(uri))

                uri = parts[0] + '/resources'
                # Make sure everything after additional / or ? in uri are ignored
                id = parts[1].split('/')[0].split('?')[0]

                # Unpack result and find correct entry
                data = self.get_json(uri)
                data = data['ResultSet']['Result']
                try:
                    data = next(x for x in data if x.get('xnat_abstractresource_id', None) == id or x.get('label', None) == id)
                except StopIteration:
                    raise XNATValueError('Could not find data for resource with abstract resource id or label matching {}'.format(id))

                # Pack data properly for xnat response
                data = {
                    'items': [
                        {
                            'children': [],
                            'meta': {
                                'xsi:type': 'xnat:resourceCatalog',
                                'isHistory': False
                            },
                            'data_fields': data
                        }
                    ]
                }

                return data
            else:
                raise XNATValueError('Could not decode JSON from [{}] {}'.format(uri, response.text))

    def download_stream(self, uri, target_stream, format=None, verbose=False, chunk_size=524288, update_func=None, timeout=None):
        """
        Download the given ``uri`` to the given ``target_stream``.

        :param str uri:            Path of the uri to retrieve.
        :param file target_stream: A writable file-like object to save the
                                   stream to.
        :param str format:         Request format
        :param bool verbose:       If ``True``, and an ``update_func`` is not
                                   specified, a progress bar is shown on
                                   stdout.
        :param int chunk_size:     Download this many bytes at a time
        :param func update_func:   If provided, will be called every
                                   ``chunk_size`` bytes. Must accept three
                                   parameters:

                                     - the number of bytes downloaded so far
                                     - the total number of bytes to be
                                       downloaded (might be ``None``),
                                     - A boolean flag which is ``False`` during
                                       the download, and ``True`` when the
                                       download has completed (or failed)
        :param timeout: timeout in seconds, float or (connection timeout, read timeout)
        :type timeout: float or tuple
        """
        self._check_connection()

        uri = self._format_uri(uri, format=format)
        self.logger.debug('DOWNLOAD STREAM {}'.format(uri))

        # Stream the get and write to file
        response = self.interface.get(uri, stream=True, timeout=timeout)

        if response.status_code not in self.accepted_status_get:
            raise exceptions.XNATResponseError('Invalid response from XNATSession for url {} (status {}):\n{}'.format(uri, response.status_code, response.text))

        # Get the content length if available
        content_length = response.headers.get('Content-Length', None)

        if isinstance(content_length, six.string_types):
            content_length = int(content_length)

        if verbose and update_func is None:
            update_func = default_update_func(content_length)
        if update_func is None:
            update_func = lambda *args: None

        if verbose:
            self.logger.info('Downloading {}:'.format(uri))

        bytes_read = 0
        try:
            update_func(0, content_length, False)
            for chunk in response.iter_content(chunk_size):
                if bytes_read == 0 and chunk[0] == '<' and chunk.startswith(('<!DOCTYPE', '<html>')):
                    raise ValueError('Invalid response from XNATSession (status {}):\n{}'.format(response.status_code, chunk))

                bytes_read += len(chunk)
                target_stream.write(chunk)

                update_func(bytes_read, content_length, False)
        finally:
            update_func(bytes_read, content_length, True)

    def download(self, uri, target, format=None, verbose=True, timeout=None):
        """
        Download uri to a target file
        """
        self._check_connection()

        with open(target, 'wb') as out_fh:
            self.download_stream(uri, out_fh, format=format, verbose=verbose, timeout=timeout)

        if verbose:
            self.logger.info('\nSaved as {}...'.format(target))

    def download_zip(self, uri, target, verbose=True, timeout=None):
        """
        Download uri to a target zip file
        """
        self.download(uri, target, format='zip', verbose=verbose, timeout=timeout)

    def upload(self, uri, file_, retries=1, query=None, content_type=None, method='put', overwrite=False, timeout=None):
        """
        Upload data or a file to XNAT

        :param str uri: uri to upload to
        :param file_: the file handle, path to a file or a string of data
                      (which should not be the path to an existing file!)
        :param int retries: amount of times xnatpy should retry in case of
                            failure
        :param dict query: extra query string content
        :param content_type: the content type of the file, if not given it will
                             default to ``application/octet-stream``
        :param str method: either ``put`` (default) or ``post``
        :param bool overwrite: indicate if previous data should be overwritten
        :param timeout: timeout in seconds, float or (connection timeout, read timeout)
        :type timeout: float or tuple
        :return:
        """
        self._check_connection()

        if overwrite:
            if query is None:
                query = {}
            query['overwrite'] = 'true'

        uri = self._format_uri(uri, query=query)
        self.logger.debug('UPLOAD URI {}'.format(uri))
        attempt = 0
        file_handle = None
        opened_file = False

        try:
            while attempt < retries:
                if isinstance(file_, FILE_TYPES):
                    # File is open file handle, seek to 0
                    file_handle = file_
                    file_.seek(0)
                # Make sure conditions are valid for os.path.isfile to function
                elif isinstance(file_, six.string_types) and '\0' not in file_ and os.path.isfile(file_):
                    # File is str path to file
                    file_handle = open(file_, 'rb')
                    opened_file = True
                else:
                    # File is data to upload
                    file_handle = file_

                attempt += 1

                try:
                    # Set the content type header
                    if content_type is None:
                        headers = {'Content-Type': 'application/octet-stream'}
                    else:
                        headers = {'Content-Type': content_type}

                    if method == 'put':
                        response = self.interface.put(uri, data=file_handle, headers=headers, timeout=timeout)
                    elif method == 'post':
                        response = self.interface.post(uri, data=file_handle, headers=headers, timeout=timeout)
                    else:
                        raise ValueError('Invalid upload method "{}" should be either put or post.'.format(method))
                    self._check_response(response)
                    return response
                except exceptions.XNATResponseError:
                    pass
        finally:
            if opened_file:
                file_handle.close()

        # We didn't return correctly, so we have an error
        raise exceptions.XNATUploadError('Upload failed after {} attempts! Status code {}, response text {}'.format(retries, response.status_code, response.text))

    @property
    def scanners(self):
        """
        A list of scanners referenced in XNATSession
        """
        return [x['scanner'] for x in self.xnat_session.get_json('/data/archive/scanners')['ResultSet']['Result']]

    @property
    def scan_types(self):
        """
         A list of scan types associated with this XNATSession instance
        """
        return self.xnat_session.get_json('/data/archive/scan_types')['ResultSet']['Result']

    @property
    @caching
    def xnat_version(self):
        """
        The version of the XNAT server
        """
        try:
            # XNAT SERVER 1.6.x
            return self.get('/data/version').text
        except exceptions.XNATResponseError:
            # XNAT SERVER 1.7.x
            return self.get_json('/xapi/siteConfig/buildInfo')['version']

    def create_object(self, uri, type_=None, fieldname=None, **kwargs):
        if (uri, fieldname) not in self._cache['__objects__']:
            if type_ is None:
                if self.xnat_session.debug:
                    self.logger.debug('Type unknown, fetching data to get type')
                data = self.xnat_session.get_json(uri)
                type_ = data['items'][0]['meta']['xsi:type']
                datafields = data['items'][0]['data_fields']
            else:
                datafields = None

            if self.xnat_session.debug:
                self.logger.debug('Looking up type {} [{}]'.format(type_, type(type_).__name__))
            if type_ not in self.XNAT_CLASS_LOOKUP:
                raise KeyError('Type {} unknow to this XNATSession REST client (see XNAT_CLASS_LOOKUP class variable)'.format(type_))

            cls = self.XNAT_CLASS_LOOKUP[type_]

            if self.xnat_session.debug:
                self.logger.debug('Creating object of type {}'.format(cls))

            # Add project post-hoc hook for fixing some problems with shared
            # resources, the .+? is the non greedy version of .+
            match = re.search('/data(?:/archive)?/projects/(.+?)/', uri)

            if match:
                # Set overwrite field
                overwrites = {'project': match.group(1)}
            else:
                overwrites = None

            obj = cls(uri, self, datafields=datafields, fieldname=fieldname, overwrites=overwrites, **kwargs)

            self._cache['__objects__'][uri, fieldname] = obj
        elif self.debug:
            self.logger.debug('Fetching object {} from cache'.format(uri))

        return self._cache['__objects__'][uri, fieldname]

    @property
    @caching
    def projects(self):
        """
        Listing of all projects on the XNAT server

        Returns an :py:class:`XNATListing <xnat.core.XNATListing>` with elements
        of :py:class:`ProjectData <xnat.classes.ProjectData>`
        """
        return XNATListing(self.uri + '/projects',
                           xnat_session=self.xnat_session,
                           parent=self,
                           field_name='projects',
                           xsi_type='xnat:projectData',
                           secondary_lookup_field='name')

    @property
    @caching
    def subjects(self):
        """
        Listing of all subjects on the XNAT server

        Returns an :py:class:`XNATListing <xnat.core.XNATListing>` with elements
        of :py:class:`SubjectData <xnat.classes.SubjectData>`
        """
        return XNATListing(self.uri + '/subjects',
                           xnat_session=self.xnat_session,
                           parent=self,
                           field_name='subjects',
                           xsi_type='xnat:subjectData',
                           secondary_lookup_field='label')

    @property
    @caching
    def experiments(self):
        """
        Listing of all experiments on the XNAT server

        Returns an :py:class:`XNATListing <xnat.core.XNATListing>` with elements
        that are subclasses of :py:class:`ExperimentData <xnat.classes.ExperimentData>`
        """
        return XNATListing(self.uri + '/experiments',
                           xnat_session=self.xnat_session,
                           parent=self,
                           field_name='experiments',
                           secondary_lookup_field='label')

    @property
    def prearchive(self):
        """
        Representation of the prearchive on the XNAT server, see :py:mod:`xnat.prearchive`
        """
        return self._prearchive

    @property
    def users(self):
        """
        Representation of the users registered on the XNAT server
        """
        return self._users

    @property
    def services(self):
        """
        Collection of services, see :py:mod:`xnat.services`
        """
        return self._services

    def clearcache(self):
        """
        Clear the cache of the listings in the Session object
        """
        self._cache.clear()
        self._cache['__objects__'] = {}


def default_update_func(total):
    """
    Set up a default update function to be used by the
    :class:`Session.download_stream` method. This function configures a
    ``progressbar.ProgressBar`` object which displays progress as a file
    is downloaded.

    :param int total: Total number of bytes to be downloaded (might be
                      ``None``)

    :returns: A function to be used as the ``update_func`` by the
              ``Session.download_stream`` method.
    """

    if total is not None:
        widgets = [
            Percentage(),
            ' of ', DataSize('max_value'),
            ' ', Bar(),
            ' ', AdaptiveTransferSpeed(),
            ' ', AdaptiveETA(),
        ]
    else:
        total = UnknownLength
        widgets = [
            DataSize(),
            ' ', BouncingBar(),
            ' ', AdaptiveTransferSpeed(),
        ]

    progress_bar = ProgressBar(widgets=widgets, max_value=total)

    # The real update function which gets called by download_stream
    def do_update(nbytes, total, finished, progress_bar=progress_bar):

        if nbytes == 0:
            progress_bar.start()
        elif finished:
            progress_bar.finish()
        else:
            progress_bar.update(nbytes)

    return do_update



# Copyright 2011-2015 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import unicode_literals

import re
import keyword
from functools import update_wrapper

import six

if six.PY3:
    from io import BufferedIOBase, SEEK_SET, SEEK_END
    superclass = BufferedIOBase
else:
    from os import SEEK_SET, SEEK_END
    superclass = object


class mixedproperty(object):
    """
    A special property-like class that can act as a property for a class as
    well as a property for an object. These properties can have different
    function so the behaviour changes depending on whether it is called on
    the class or and instance of the class.
    """
    def __init__(self, fcget, fget=None, fset=None, fdel=None):
        # fcget is the get on the class e.g. Test.x
        # fget is the get on an instance Test().x
        # fset and fdel are the set and delete of the instance
        self.fcget = fcget
        self.fget = fget
        self.fset = fset
        self.fdel = fdel

        update_wrapper(self, fcget)

    def __get__(self, obj, objtype):
        if obj is not None and self.fget is not None:
            # If the obj is None, it is called on the class
            # If the fget is not set, call the class version
            return self.fget(obj)
        else:
            # Splice the docstring into returned object
            value = self.fcget(objtype)

            # Check if we can safely copy docstring and do so if possible
            from_module = value.__class__.__module__

            if from_module is not None and from_module.startswith('xnat.'):
                value.__doc__ = self.__doc__

            return value

    def __set__(self, obj, value):
        if self.fset is None:
            raise AttributeError("can't set attribute")
        self.fset(obj, value)

    def __delete__(self, obj):
        if self.fdel is None:
            raise AttributeError("can't delete attribute")
        self.fdel(obj)

    # These allow the updating of the property using the @x.getter, @x.setter
    # and @x.deleter decorators.
    def getter(self, fget):
        return type(self)(self.fcget, fget, self.fset, self.fdel)

    def setter(self, fset):
        return type(self)(self.fcget, self.fget, fset, self.fdel)

    def deleter(self, fdel):
        return type(self)(self.fcget, self.fget, self.fset, fdel)


def pythonize_class_name(name):
    """
    Turns string into a valid PEP8 class name, meaning camel cased
    (e.g. someValue -> SomeValue)

    :param str name: the name to convert to a PEP8 valid version
    :return: the PEP8 valid class name
    :rtype: str
    """
    if ':' in name:
        name = name.split(':', 1)[-1]

    parts = re.split('[\-\_\W]+', name)
    parts = [x[0].upper() + x[1:] for x in parts]
    name = ''.join(parts)
    return name


def pythonize_attribute_name(name):
    """
    Turns string into a valid PEP8 class name, meaning lower case with
    underscores when needed (e.g. someValue -> some_value)

    :param str name: the name to convert to a PEP8 valid version
    :return: the PEP8 valid attribute name
    :rtype: str
    """
    name = re.sub('[^0-9a-zA-Z]+', '_', name)

    # Change CamelCaseString to camel_case_string
    # Note that addID would become add_id
    name = re.sub("[A-Z]+", lambda x: '_' + x.group(0).lower(), name)
    if name[0] == '_':
        name = name[1:]

    # Avoid multiple underscores (replace them by single underscore)
    name = re.sub("__+", '_', name)

    # Avoid overwriting keywords TODO: Do we want this, as a property it is not a huge problem?
    if keyword.iskeyword(name):
        name += '_'

    return name


class RequestsFileLike(object):
    def __init__(self, request, chunk_size=512*1024):
        self._bytes = six.BytesIO()
        self._request = request
        self._iterator = request.iter_content(chunk_size)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def _load_all(self):
        self._bytes.seek(0, SEEK_END)
        for chunk in self._iterator:
            self._bytes.write(chunk)

    def _load_until(self, goal_position):
        current_position = self._bytes.seek(0, SEEK_END)
        while current_position < goal_position:
            try:
                current_position = self._bytes.write(next(self._iterator))
            except StopIteration:
                break

    def tell(self):
        return self._bytes.tell()

    def read(self, size=None):
        current_position = self._bytes.tell()

        if size is None:
            self._load_all()
        else:
            goal_position = current_position + size
            self._load_until(goal_position)

        self._bytes.seek(current_position)
        return self._bytes.read(size)

    def seek(self, position, whence=SEEK_SET):
        if whence == SEEK_END:
            self._load_all()
        else:
            self._bytes.seek(position, whence)

    def close(self):
        self._bytes.close()
        self._request.close()


def full_class_name(cls):
    module = cls.__module__

    if module is None or module == str.__module__:
        return cls.__name__

    return '{cls.__module__}.{cls.__name__}'.format(cls=cls)



# Copyright 2011-2015 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from six.collections_abc import Mapping
from .core import caching


class Users(Mapping):
    """
    Listing of the users on the connected XNAT installation
    """
    def __init__(self, xnat_session):
        # cache fields
        self._cache = {}
        self.caching = True

        # keep session available
        self._xnat_session = xnat_session

    def __repr__(self):
        return '<Users {}>'.format(self.data)

    def __getitem__(self, item):
        return self.data[item]

    def __len__(self):
        return len(self.data)

    def __iter__(self):
        for x in self.data:
            yield x

    @property
    def xnat_session(self):
        return self._xnat_session

    @property
    @caching
    def data(self):
        users = self.xnat_session.get_json('/data/users')['ResultSet']['Result']
        return {x['login']: User(x) for x in users}


class User(object):
    """
    Representation of a user on the connected XNAT systen
    """
    def __init__(self, data):
        self._fulldata = data

    def __repr__(self):
        return '<User {} [{}]>'.format(self.login, self.id)

    @property
    def data(self):
        return self._fulldata

    @property
    def id(self):
        """
        The id of the user
        """
        return self.data['xdat_user_id']

    @property
    def login(self):
        """
        The login name of the user
        """
        return self.data['login']

    @property
    def email(self):
        """
        The email of the user
        """
        return self.data['email']

    @property
    def first_name(self):
        """
        The first name of the user
        """
        return self.data['firstname']

    @property
    def last_name(self):
        """
        The last name of the user
        """
        return self.data['lastname']



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This package contains the entire client. The connect function is the only
function actually in the package. All following classes are created based on
the https://central.xnat.org/schema/xnat/xnat.xsd schema and the xnatcore and
xnatbase modules, using the convert_xsd.
"""

from __future__ import absolute_import
from __future__ import unicode_literals
import getpass
import hashlib
import imp
import logging
import os
import platform
import netrc
import re
import tempfile
import time

import requests
from six.moves.urllib import parse

from . import exceptions
from .session import XNATSession
from .constants import DEFAULT_SCHEMAS
from .convert_xsd import SchemaParser

GEN_MODULES = {}

__version__ = '0.3.23'
__all__ = ['connect', 'exceptions']


def check_auth(requests_session, server, user, logger):
    """
    Try to figure out of the requests session is properly logged in as the desired user

    :param requests.Session requests_session: requests session
    :param str server: server test
    :param str user: desired user (None for guest)
    :raises ValueError: Raises a ValueError if the login failed
    """
    logger.debug('Getting {} to test auth'.format(server))
    test_auth_request = requests_session.get(server)
    logger.debug('Status: {}'.format(test_auth_request.status_code))

    if test_auth_request.status_code == 401 or 'Login attempt failed. Please try again.' in test_auth_request.text:
        message = 'Login attempt failed for {}, please make sure your credentials for user {} are correct!'.format(server, user)
        logger.critical(message)
        raise exceptions.XNATLoginFailedError(message)

    if test_auth_request.status_code != 200:
        logger.warning('Simple test requests did not return a 200 code! Server might not be functional!')

    if user is not None:
        match = re.search(r'<span id="user_info">Logged in as: &nbsp;<a (id="[^"]+" )?href="[^"]+">(?P<username>[^<]+)</a>',
                          test_auth_request.text)

        if match is None:
            match = re.search(r'<span id="user_info">Logged in as: <span style="color:red;">Guest</span>',
                              test_auth_request.text)
            if match is None:
                match = re.search('Your password has expired', test_auth_request.text)
                if match:
                    message = 'Your password has expired. Please try again after updating your password on XNAT.'
                    logger.error(message)
                    raise exceptions.XNATExpiredCredentialsError(message)
                else:
                    message = 'Could not determine if login was successful!'
                    logger.error(message)
                    logger.debug(test_auth_request.text)
                    raise exceptions.XNATAuthError(message)
            else:
                message = 'Login failed (in guest mode)!'
                logger.error(message)
                raise exceptions.XNATLoginFailedError(message)
        else:
            username = match.group('username')
            logger.info('Logged in successfully as {}'.format(username))
            return username

    match = re.search(r'<span id="user_info">Logged in as: <span style="color:red;">Guest</span>',
                      test_auth_request.text)
    if match is None:
        message = 'Could not determine if login was successful!'
        logger.error(message)
        raise exceptions.XNATAuthError(message)
    else:
        logger.info('Logged in as guest successfully')
        return 'guest'


def parse_schemas_16(parser, xnat_session, extension_types=True):
    """
    Retrieve and parse schemas for an XNAT version 1.6.x

    :param parser: The parser to use for the parsing
    :param requests_session: the requests session used for the communication
    :param str server: URI of the XNAT server
    :param logger: logger to use for the logging
    :param bool extension_types: flag to enabled/disable scanning for extension types
    """
    # Retrieve schema from XNAT server
    schema_uri = '/schemas/xnat/xnat.xsd'

    success = parser.parse_schema_uri(xnat_session=xnat_session,
                                      schema_uri=schema_uri)

    if not success:
        raise RuntimeError('Could not parse the xnat.xsd! See error log for details!')

    # Parse extension types
    if extension_types:
        projects_uri = '/data/projects?format=json'
        try:
            response = xnat_session.get(projects_uri)
        except exceptions.XNATResponseError as exception:
            message = 'Could list projects while scanning for extension types: {}'.format(exception)
            xnat_session.logger.critical(message)
            raise exception

        try:
            project_id = response.json()['ResultSet']['Result'][0]['ID']
        except (KeyError, IndexError):
            raise ValueError('Could not find an example project for scanning extension types!')

        project_uri = '/data/projects/{}'.format(project_id)
        try:
            response = xnat_session.get(project_uri, format='xml')
        except exceptions.XNATResponseError as exception:
            message = 'Could load example project while scanning for extension types: {}'.format(exception)
            xnat_session.logger.critical(message)
            raise exception

        schemas = parser.find_schema_uris(response.text)
        if schema_uri in schemas:
            xnat_session.logger.debug('Removing schema {} from list'.format(schema_uri))
            schemas.remove(schema_uri)
        xnat_session.logger.info('Found additional schemas: {}'.format(schemas))

        for schema in schemas:
            parser.parse_schema_uri(xnat_session=xnat_session,
                                    schema_uri=schema)


def parse_schemas_17(parser, xnat_serssion, extension_types=True):
    """
    Retrieve and parse schemas for an XNAT version 1.7.x

    :param parser: The parser to use for the parsing
    :param requests_session: the requests session used for the communication
    :param str server: URI of the XNAT server
    :param logger: logger to use for the logging
    """
    if extension_types:
        schemas_uri = '/xapi/schemas'
        try:
            schema_list = xnat_serssion.get_json(schemas_uri)
        except exceptions.XNATResponseError as exception:
            message = 'Problem retrieving schemas list: {}'.format(exception)
            xnat_serssion.logger.critical(message)
            raise ValueError(message)
    else:
        schema_list = DEFAULT_SCHEMAS

    for schema in schema_list:
        if extension_types or schema in ['xdat', 'xnat']:
            parser.parse_schema_uri(xnat_session=xnat_serssion,
                                    schema_uri='/xapi/schemas/{schema}'.format(schema=schema))


def detect_redirection(server, session, logger):
    """
    Check if there is a redirect going on
    :param str server: server url
    :param session: requests session to use for the connection
    :param logger:
    :return: the server url to use later
    """
    response = session.get(server.rstrip('/') + '/data/projects')
    logger.debug('Response url: {}'.format(response.url))
    response_url = re.match(r'(.*)/(app|data)', response.url).group(1)
    if response_url != server and response_url != server + '/':
        logger.warning('Detected a redirect from {0} to {1}, using {1} from now on'.format(server, response_url))
    return response_url


def query_netrc(server, netrc_file, logger):
    # Get the login info
    parsed_server = parse.urlparse(server)

    logger.info('Retrieving login info for {}'.format(parsed_server.netloc))
    try:
        if netrc_file is None:
            netrc_file = os.path.join('~', '_netrc' if os.name == 'nt' else '.netrc')
            netrc_file = os.path.expanduser(netrc_file)
        user, _, password = netrc.netrc(netrc_file).authenticators(parsed_server.netloc)
        logger.info('Found login for {}'.format(parsed_server.netloc))
    except (TypeError, IOError):
        logger.info('Could not find login for {}, continuing without login'.format(parsed_server.netloc))
        user = password = None

    return user, password


def build_model(xnat_session, extension_types, connection_id):
    """
    Build the XNAT data model for a given connection
    """
    logger = xnat_session.logger
    debug = xnat_session.debug

    # Check XNAT version
    version = xnat_session.xnat_version

    # Generate module
    parser = SchemaParser(debug=debug, logger=logger)

    if xnat_session.xnat_version.startswith('1.6'):
        logger.info('Found an 1.6 version ({})'.format(version))
        parse_schemas_16(parser, xnat_session, extension_types=extension_types)
    elif version.startswith('1.7'):
        logger.info('Found an 1.7 version ({})'.format(version))
        parse_schemas_17(parser, xnat_session, extension_types=extension_types)
    else:
        logger.critical('Found an unsupported version ({})'.format(version))
        raise ValueError('Cannot continue on unsupported XNAT version')

    # Write code to temp file
    with tempfile.NamedTemporaryFile(mode='w', suffix='_generated_xnat.py', delete=False) as code_file:
        parser.write(code_file=code_file)

    logger.debug('Code file written to: {}'.format(code_file.name))

    # The module is loaded in its private namespace based on the code_file name
    xnat_module = imp.load_source('xnat_gen_{}'.format(connection_id),
                                  code_file.name)
    xnat_module._SOURCE_CODE_FILE = code_file.name

    logger.debug('Loaded generated module')

    # Register all types parsed
    for cls in parser.class_list.values():
        if not (cls.name is None or (cls.base_class is not None and cls.base_class.startswith('xs:'))):
            getattr(xnat_module, cls.writer.python_name).__register__(xnat_module.XNAT_CLASS_LOOKUP)

    xnat_module.SESSION = xnat_session

    # Add the required information from the module into the xnat_session object
    xnat_session.XNAT_CLASS_LOOKUP.update(xnat_module.XNAT_CLASS_LOOKUP)
    xnat_session.classes = xnat_module
    xnat_session._source_code_file = code_file.name


def connect(server, user=None, password=None, verify=True, netrc_file=None, debug=False,
            extension_types=True, loglevel=None, logger=None, detect_redirect=True,
            no_parse_model=False, default_timeout=300):
    """
    Connect to a server and generate the correct classed based on the servers xnat.xsd
    This function returns an object that can be used as a context operator. It will call
    disconnect automatically when the context is left. If it is used as a function, then
    the user should call ``.disconnect()`` to destroy the session and temporary code file.

    :param str server: uri of the server to connect to (including http:// or https://)
    :param str user: username to use, leave empty to use netrc entry or anonymous login.
    :param str password: password to use with the username, leave empty when using netrc.
                         If a username is given and no password, there will be a prompt
                         on the console requesting the password.
    :param bool verify: verify the https certificates, if this is false the connection will
                        be encrypted with ssl, but the certificates are not checked. This is
                        potentially dangerous, but required for self-signed certificates.
    :param str netrc_file: alternative location to use for the netrc file (path pointing to
                           a file following the netrc syntax)
    :param debug bool: Set debug information printing on and print extra debug information.
                       This is meant for xnatpy developers and not for normal users. If you
                       want to debug your code using xnatpy, just set the loglevel to DEBUG
                       which will show you all requests being made, but spare you the
                       xnatpy internals.
    :param str loglevel: Set the level of the logger to desired level
    :param logging.Logger logger: A logger to reuse instead of creating an own logger
    :param bool detect_redirect: Try to detect a redirect (via a 302 response) and short-cut
                                 for subsequent requests
    :param bool no_parse_model: Create an XNAT connection without parsing the server data
                                model, this create a connection for which the simple
                                get/head/put/post/delete functions where, but anything
                                requiring the data model will file (e.g. any wrapped classes)
    :param int default_timeout: The default timeout of requests sent by xnatpy, is a 5 minutes
                                per default.
    :return: XNAT session object
    :rtype: XNATSession

    Preferred use::

        >>> import xnat
        >>> with xnat.connect('https://central.xnat.org') as session:
        ...    subjects = session.projects['Sample_DICOM'].subjects
        ...    print('Subjects in the SampleDICOM project: {}'.format(subjects))
        Subjects in the SampleDICOM project: <XNATListing (CENTRAL_S01894, dcmtest1): <SubjectData CENTRAL_S01894>, (CENTRAL_S00461, PACE_HF_SUPINE): <SubjectData CENTRAL_S00461>>

    Alternative use::

        >>> import xnat
        >>> session = xnat.connect('https://central.xnat.org')
        >>> subjects = session.projects['Sample_DICOM'].subjects
        >>> print('Subjects in the SampleDICOM project: {}'.format(subjects))
        Subjects in the SampleDICOM project: <XNATListing (CENTRAL_S01894, dcmtest1): <SubjectData CENTRAL_S01894>, (CENTRAL_S00461, PACE_HF_SUPINE): <SubjectData CENTRAL_S00461>>
        >>> session.disconnect()
    """

    # Generate a hash for the connection
    hasher = hashlib.md5()
    hasher.update(server.encode('utf-8'))
    hasher.update(str(time.time()).encode('utf-8'))
    connection_id = hasher.hexdigest()

    # Setup the logger for this connection
    if logger is None:
        logger = logging.getLogger('xnat-{}'.format(connection_id))
        handler = logging.StreamHandler()
        handler.setLevel(logging.DEBUG)
        logger.addHandler(handler)

        # create formatter
        if debug:
            formatter = logging.Formatter('[%(asctime)s] %(levelname)s: %(module)s:%(lineno)d >> %(message)s')
        else:
            formatter = logging.Formatter('[%(levelname)s] %(message)s')
        handler.setFormatter(formatter)

        if loglevel is not None:
            logger.setLevel(loglevel)
        elif debug:
            logger.setLevel('DEBUG')
        else:
            logger.setLevel('WARNING')

    # If verify is False, disable urllib3 warning and give a one time warning!
    if not verify:
        logger.warning('Verify is disabled, this will NOT verify the certificate of SSL connections!')
        logger.warning('Warnings about invalid certificates will be HIDDEN to avoid spam, but this')
        logger.warning('means that your connection can be potentially unsafe!')
        requests.packages.urllib3.disable_warnings()

    if user is None and password is None:
        user, password = query_netrc(server, netrc_file, logger)

    if user is not None and password is None:
        password = getpass.getpass(prompt=str("Please enter the password for user '{}':".format(user)))

    # Create the correct requests session
    requests_session = requests.Session()
    user_agent = "xnatpy/{version} ({platform}/{release}; python/{python}; requests/{requests})".format(
        version=__version__,
        platform=platform.system(),
        release=platform.release(),
        python=platform.python_version(),
        requests=requests.__version__
    )

    requests_session.headers.update({'User-Agent': user_agent})

    if user is not None:
        requests_session.auth = (user, password)

    if not verify:
        requests_session.verify = False

    # Check for redirects
    original_uri = server
    if detect_redirect:
        server = detect_redirection(server, requests_session, logger)

    # If no login was found, check if the new server has a known login as a backup
    if original_uri != server and user is None and password is None:
        user, password = query_netrc(server, netrc_file, logger)

    # Check if login is successful
    logged_in_user = check_auth(requests_session, server=server, user=user, logger=logger)

    # Create the XNAT connection
    xnat_session = XNATSession(server=server, logger=logger,
                               interface=requests_session, debug=debug,
                               original_uri=original_uri, logged_in_user=logged_in_user,
                               default_timeout=default_timeout)

    # Parse data model and create classes
    if not no_parse_model:
        build_model(xnat_session, extension_types=extension_types, connection_id=connection_id)

    return xnat_session



# Copyright 2011-2015 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import unicode_literals
import fnmatch


class Inspect(object):
    def __init__(self, xnat_session):
        self._xnat_session = xnat_session

    @property
    def xnat_session(self):
        return self._xnat_session

    def datatypes(self, pattern='*', fields_pattern=None):
        elements = self.xnat_session.get_json('/data/search/elements')

        elements = [x['ELEMENT_NAME'] for x in elements['ResultSet']['Result']]

        # Filter fields using pattern
        if '*' in pattern or '?' in pattern:
            elements = [field for field in elements if fnmatch.fnmatch(field, pattern)]

        if fields_pattern is None:
            return elements
        else:
            return [field for element in elements for field in self.datafields(datatype=element, pattern=fields_pattern)]

    def datafields(self, datatype, pattern='*', prepend_type=True):
        search_fields = self.xnat_session.get_json('/data/search/elements/{}'.format(datatype))

        # Select data from JSON
        search_fields = [x['FIELD_ID'] for x in search_fields['ResultSet']['Result']]

        # Filter fields using pattern
        if '*' in pattern or '?' in pattern:
            search_fields = [field for field in search_fields if fnmatch.fnmatch(field, pattern)]

        # Filter fields for unwanted fields
        search_fields = [field for field in search_fields if '=' not in field and 'SHARINGSHAREPROJECT' not in field]

        return ['{}/{}'.format(datatype, field) if prepend_type else field for field in search_fields]




# Copyright 2011-2015 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import unicode_literals
import mimetypes
import collections
import os
import tempfile
from zipfile import ZipFile

from six import BytesIO

from .prearchive import PrearchiveSession
from .exceptions import XNATResponseError, XNATValueError

TokenResult = collections.namedtuple('TokenResult', ('alias', 'secret'))


class Services(object):
    """
    The class representing all service functions in XNAT found in the
    /data/services REST directory
    """
    def __init__(self, xnat_session):
        self._xnat_session = xnat_session

    @property
    def xnat_session(self):
        return self._xnat_session

    def dicom_dump(self, src, fields=None):
        """
        Retrieve a dicom dump as a JSON data structure
        See the XAPI documentation for more detailed information: `DICOM Dump Service <https://wiki.xnat.org/display/XAPI/DICOM+Dump+Service+API>`_

        :param str src: The url of the scan to generate the DICOM dump for
        :param list fields: Fields to filter for DICOM tags. It can either a tag name or tag number in the format GGGGEEEE (G = Group number, E = Element number)
        :return: JSON object (dict) representation of DICOM header
        :rtype: dict
        """
        query_string = {'src': src}
        if fields is not None:
            if not isinstance(fields, (list, str)):
                raise XNATValueError('The fields argument to .dicom_dump() should be list or a str and not {}'.format(type(fields)))
            query_string['field'] = fields

        return self.xnat_session.get_json('/data/services/dicomdump', query=query_string)['ResultSet']['Result']

    def import_(self, path, overwrite=None, quarantine=False, destination=None,
                trigger_pipelines=None, project=None, subject=None,
                experiment=None, content_type=None, import_handler=None):
        """
        Import a file into XNAT using the import service. See the
        `XNAT wiki <https://wiki.xnat.org/pages/viewpage.action?pageId=6226268>`_
        for a detailed explanation.

        :param str path: local path of the file to upload and import
        :param str overwrite: how the handle existing data (none, append, delete)
        :param bool quarantine: flag to indicate session should be quarantined
        :param bool trigger_pipelines: indicate that archiving should trigger pipelines
        :param str destination: the destination to upload the scan to
        :param str project: the project in the archive to assign the session to
                            (only accepts project ID, not a label)
        :param str subject: the subject in the archive to assign the session to
        :param str experiment: the experiment in the archive to assign the session content to
        :param str content_type: overwite the content_type (by default the mimetype will be
                                 guessed using the ``mimetypes`` package).
                                 This will often be ``application/zip``.
        :return:

        .. note::
            The project has to be given using the project ID and *NOT* the label.

        .. warning::
            On some systems the guessed mimetype of a zip file might not be ``application/zip``
            but be something like ``application/x-zip-compressed``. In that case you might have to
            set the ``content_type`` parameter to ``application/zip`` manually.

        """
        query = {}
        if overwrite is not None:
            if overwrite not in ['none', 'append', 'delete']:
                raise ValueError('Overwrite should be none, append or delete!')
            query['overwrite'] = overwrite

        if quarantine:
            query['quarantine'] = 'true'

        if trigger_pipelines is not None:
            if isinstance(trigger_pipelines, bool):
                if trigger_pipelines:
                    query['triggerPipelines'] = 'true'
                else:
                    query['triggerPipelines'] = 'false'
            else:
                raise TypeError('trigger_pipelines should be a boolean')

        if destination is not None:
            query['dest'] = destination

        if project is not None:
            query['project'] = project

        if subject is not None:
            query['subject'] = subject

        if experiment is not None:
            query['session'] = experiment

        if import_handler is not None:
            query['import-handler'] = import_handler

        # Get mimetype of file
        if content_type is None:
            content_type, _ = mimetypes.guess_type(path)

        if content_type not in ['application/x-tar', 'application/zip']:
            if path.endswith('.zip'):
                self.xnat_session.logger.warning(
                    'Found unexpect content type, but assuming zip based on the extension'
                )
                content_type = 'application/zip'
            elif path.endswith('.tar.gz'):
                self.xnat_session.logger.warning(
                    'Found unexpect content type, but assuming tar based on the extension'
                )
                content_type = 'application/x-tar'
            else:
                self.xnat_session.logger.warning(
                    'Found unexpected content (found type {}), this could result in errors on import!'.format(content_type)
                )

        uri = '/data/services/import'
        response = self.xnat_session.upload(uri=uri, file_=path, query=query, content_type=content_type, method='post')

        if response.status_code != 200:
            raise XNATResponseError('The response for uploading was ({}) {}'.format(response.status_code, response.text))

        # Create object, the return text should be the url, but it will have a \r\n at the end that needs to be stripped
        response_text = response.text.strip()
        if response_text.startswith('/data/prearchive'):
            return PrearchiveSession(response_text, self.xnat_session)

        return self.xnat_session.create_object(response_text)

    def import_dir(self, directory, overwrite=None, quarantine=False, destination=None,
                trigger_pipelines=None, project=None, subject=None,
                experiment=None, method='zip_file', import_handler=None):
        """
        Import a directory to an XNAT resource.

        :param str path: local path of the directory to upload and import
        :param str overwrite: how the handle existing data (none, append, delete)
        :param bool quarantine: flag to indicate session should be quarantined
        :param bool trigger_pipelines: indicate that archiving should trigger pipelines
        :param str destination: the destination to upload the scan to
        :param str project: the project in the archive to assign the session to
                            (only accepts project ID, not a label)
        :param str subject: the subject in the archive to assign the session to
        :param str experiment: the experiment in the archive to assign the session content to


        The method has 2 options, default is zip_file:

        #. ``zip_file``: Create a temporary zip file and upload that
        #. ``zip_file``: Create a temporary zip file in memory and upload it

        The considerations are that sometimes you can fit things in memory so
        you can save disk IO by putting it in memory.

        """
        # Make sure that a None or empty string is replaced by the default
        method = method or 'zip_file'

        if method == 'zip_file':
            content_type = 'application/zip'
            fh = tempfile.TemporaryFile('wb+')
            with ZipFile(fh, 'w') as zip_file:
                for dirpath, dirs, files in os.walk(directory):
                    for f in files:
                        fn = os.path.join(dirpath, f)
                        zip_file.write(fn)
        elif method == 'zip_memory':
            content_type = 'application/zip'
            fh = BytesIO()
            with ZipFile(fh, 'w') as zip_file:
                for dirpath, dirs, files in os.walk(directory):
                    for f in files:
                        fn = os.path.join(dirpath, f)
                        zip_file.write(fn)

        else:
            print('Selected invalid import directory method!')
            return
        fh.seek(0)
        session = self.import_(fh, overwrite, quarantine, destination,
                trigger_pipelines, project, subject,
                experiment, content_type, import_handler=import_handler)
        fh.close()
        return session



    def issue_token(self, user=None):
        """
        Issue a login token, by default for the current logged in user. If
        username is given, for that user. To issue tokens for other users
        you must be an admin.

        :param str user: User to issue token for, default is current user
        :return: Token in a named tuple (alias, secret)
        """
        uri = '/data/services/tokens/issue'
        if user:
            uri += '/user/{}'.format(user)

        result = self.xnat_session.get_json(uri)

        return TokenResult(result['alias'], result['secret'])



# Copyright 2011-2015 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals
from abc import ABCMeta, abstractmethod, abstractproperty
import collections
import contextlib
import inspect
import keyword
import os
import re
from xml.etree import ElementTree

from . import core
from . import mixin
from .datatypes import TYPE_TO_PYTHON
from .constants import SECONDARY_LOOKUP_FIELDS, FIELD_HINTS, CORE_REST_OBJECTS
from .utils import pythonize_class_name, pythonize_attribute_name, full_class_name


FILE_HEADER = '''
# Copyright 2011-2015 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import unicode_literals
import os
import tempfile  # Needed by generated code
from gzip import GzipFile  # Needed by generated code
from tarfile import TarFile  # Needed by generated code
from zipfile import ZipFile  # Needed by generated code
from six import BytesIO  # Needed by generated code

from xnat import search, mixin
from xnat.core import XNATObject, XNATNestedObject, XNATSubObject, XNATListing, XNATSimpleListing, XNATSubListing, caching
from xnat.utils import mixedproperty, RequestsFileLike

try:
    PYDICOM_LOADED = True
    import pydicom
except ImportError:
    PYDICOM_LOADED = False

SESSION = None


def current_session():
    return SESSION


# These mixins are to set the xnat_session automatically in all created classes
class XNATObjectMixin(XNATObject):
    @mixedproperty
    def xnat_session(self):
        return current_session()

    @classmethod
    def query(cls, *constraints):
        query = search.Query(cls, cls.xnat_session)

        # Add in constraints immediatly
        if len(constraints) > 0:
            query = query.filter(*constraints)

        return query


class XNATNestedObjectMixin(XNATNestedObject):
    @mixedproperty
    def xnat_session(self):
        return current_session()


class XNATSubObjectMixin(XNATSubObject):
    @mixedproperty
    def xnat_session(self):
        return current_session()


class FileData(XNATObjectMixin):
    SECONDARY_LOOKUP_FIELD = "{file_secondary_lookup}"
    _XSI_TYPE = 'xnat:fileData'

    def __init__(self, uri=None, xnat_session=None, id_=None, datafields=None, parent=None, fieldname=None, overwrites=None, path=None):
        super(FileData, self).__init__(uri=uri,
                                       xnat_session=xnat_session,
                                       id_=id_,
                                       datafields=datafields,
                                       parent=parent,
                                       fieldname=fieldname,
                                       overwrites=overwrites)

        if path is not None:
            self._path = path

    @property
    def path(self):
        return self._path

    def delete(self):
        self.xnat_session.delete(self.uri)

    def download(self, *args, **kwargs):
        self.xnat_session.download(self.uri, *args, **kwargs)

    def download_stream(self, *args, **kwargs):
        self.xnat_session.download_stream(self.uri, *args, **kwargs)
        
    def open(self):
        uri = self.xnat_session._format_uri(self.uri)
        request = self.xnat_session.interface.get(uri, stream=True)
        return RequestsFileLike(request)

    @property
    @caching
    def size(self):
        response = self.xnat_session.head(self.uri, allow_redirects=True)
        return response.headers['Content-Length']


# Empty class lookup to place all new lookup values
XNAT_CLASS_LOOKUP = {{
    "xnat:fileData": FileData,
}}


# The following code represents the data structure of the XNAT server
# It is automatically generated using
{schemas}


'''

# TODO: Add more fields to FileData from [Name, Size, URI, cat_ID, collection, file_content, file_format, tile_tags]?
# TODO: Add display identifiers support (DONE?)
# <xs:annotation>
# <xs:appinfo>
# <xdat:element displayIdentifiers="label"/>
# </xs:appinfo>
# <xs:documentation>An individual person involved in experimental research</xs:documentation>
# </xs:annotation>
# <xs:sequence>
# TODO: Add XPATHs for setting SubObjects (SEMI-DONE)
# TODO: Make Listings without key and with numeric index possible (inProgress)
# TODO: Fix scan parameters https://groups.google.com/forum/#!topic/xnat_discussion/GBZoamC2ZmY
# TODO: Check the nesting weirdness in petScanDataParametersFramesFrames (DONE?)
# TODO: Figure out the object/subobject/semi-subobject mess.
# TODO: Move all system function to use a __ prefix


class ClassPrototype(object):
    def __init__(self, parser, name, logger, field_name=None, parent_class=None, simple=False):
        self.parser = parser
        self.name = name  # This is the XSI type
        self.logger = logger

        self.field_name = field_name
        self.parent_class = parent_class
        self.parent_property = None

        self.attributes = collections.OrderedDict()
        self.base_class = None
        self.display_identifier = None
        self.abstract = False
        self._simple = simple

        self._writer = None
        self.source_schema = self.parser.current_schema

    def __repr__(self):
        return "<ClassPrototype {}>".format(self.name)

    def root_base_class(self, topxsd=False):
        base = self.base_class
        if self.base_class is None:
            return self.name

        while not base.startswith('XNAT'):
            try:
                cls = self.parser.class_list[base]
            except KeyError:
                self.logger.debug('Class list: {}'.format(
                    self.parser.class_list.keys())
                )
                raise

            if cls.base_class is None:
                return base

            if topxsd and cls.base_class.startswith('XNAT'):
                return base

            base = cls.base_class

        return base

    @property
    def simple(self):
        root_base = self.root_base_class(topxsd=True)
        root_base = self.parser.class_list.get(root_base, self)
        if root_base is self:
            return self._simple
        else:
            return root_base.simple

    @property
    def class_type(self):
        if self.simple:
            return 'Simple'
        if self.field_name is not None:
            return 'SubObject'
        elif self.name in CORE_REST_OBJECTS:
            return 'Object'
        else:
            if self.base_class is not None:
                root_base = self.root_base_class(topxsd=True)
                return self.parser.class_list[root_base].class_type
            else:
                return 'NestedObject'

    @property
    def writer(self):
        if self._writer is None:
            writers = {
                'Object': ObjectClassWriter,
                'SubObject': SubObjectClassWriter,
                'NestedObject': NestedObjectClassWriter,
                'Simple': SimpleClassWriter,
            }

            self._writer = writers[self.class_type](self)

        return self._writer

    def tostring(self):
        """
        Generate a string of the code for this class. We do this by finding the
        appropriate writer (depending on the type of class we are generating)
        and then delegating the actual generation to the writer.
        """
        return self.writer.tostring()


class AttributePrototype(object):
    def __init__(self, parser, name, logger, type=None, parent_class=None, parent_property=None,
                 min_occur=None, max_occur=None):
        self.parser = parser
        self.name = name
        self.logger = logger
        self.type = type
        self.parent_class = parent_class
        self.parent_property = parent_property
        self.field_name = None

        # Set initial values for possible additions
        self.restrictions = collections.OrderedDict()
        self.element_class = None
        self.min_occur = min_occur
        self.max_occur = max_occur
        self.docstring = None
        self.display_identifier = None

        self._writer = None

    @property
    def clean_name(self):
        return pythonize_attribute_name(self.name)

    @property
    def property_type(self):
        if self.max_occur is not None and (self.max_occur == 'unbounded' or int(self.max_occur) > 1):
            return 'listing'
        if isinstance(self.type, str):
            if self.type.startswith('xs:'):
                return 'property'
            else:
                return 'subobject'
        else:
            return 'subobject'

    @property
    def writer(self):
        if self._writer is None:
            writers = {
                'constant': ConstantWriter,
                'listing': ListingPropertyWriter,
                'property': PropertyWriter,
                'subobject': SubObjectPropertyWriter,
            }

            self._writer = writers[self.property_type](self)

        return self._writer

    def tostring(self):
        """
        Generate a string of the code for this property. We do this by finding the
        appropriate writer (depending on the type of property we are generating)
        and then delegating the actual generation to the writer.
        """
        return self.writer.tostring()

    def __repr__(self):
        return "<AttributePrototype [{}] {}>".format(self.property_type, self.name)


class BaseWriter(object):
    __metaclass__ = ABCMeta

    def __init__(self, prototype):
        self.prototype = prototype

    @abstractmethod
    def tostring(self):
        """String version"""

    @property
    def logger(self):
        return self.prototype.logger


class BaseClassWriter(BaseWriter):
    def __init__(self, prototype):
        super(BaseClassWriter, self).__init__(prototype=prototype)

    # Give easy access to prototypes attributes
    @property
    def parser(self):
        return self.prototype.parser

    @property
    def name(self):
        return self.prototype.name

    @property
    def field_name(self):
        return self.prototype.field_name

    @property
    def parent_class(self):
        return self.prototype.parent_class

    @property
    def parent_property(self):
        return self.prototype.parent_property

    @property
    def attributes(self):
        return self.prototype.attributes

    @property
    def base_class(self):
        if self.prototype.base_class is None:
            return self.default_base_class

        return self.prototype.base_class

    @property
    def display_identifier(self):
        if self.prototype.display_identifier is None and self.name in SECONDARY_LOOKUP_FIELDS:
            return SECONDARY_LOOKUP_FIELDS[self.name]
        else:
            return self.prototype.display_identifier

    @display_identifier.setter
    def display_identifier(self, value):
        self._display_identifier = value

    @property
    def abstract(self):
        return self.prototype.abstract

    @property
    def simple(self):
        return self.prototype.simple

    @property
    def source_schema(self):
        return self.prototype.source_schema

    # Convenience methods that can be shared along writers
    def get_base_template(self):
        if hasattr(mixin, self.python_name):
            return getattr(mixin, self.python_name)

    def _pythonize_name(self, name):
        return pythonize_class_name(name)

    @property
    def python_name(self):
        return pythonize_class_name(self.name)

    @property
    def python_base_class(self):
        return pythonize_class_name(self.base_class)

    @property
    def python_parent_class(self):
        return pythonize_class_name(self.parent_class)

    def hasattr(self, name):
        base = self.get_base_template()

        if base is not None:
            return hasattr(base, name)
        else:
            base = self.parser.class_list.get(self.base_class)
            if base is not None:
                if isinstance(base, ClassPrototype):
                    return base.writer.hasattr(name)
                else:
                    return base.hasattr(name)
            else:
                base = self.get_super_class()
                return hasattr(base, name)

    def get_super_class(self):
        if hasattr(core, self.python_base_class):
            return getattr(core, self.python_base_class)

    # Shared functional blocks
    def tostring(self):
        header = self.header()
        properties = '\n\n'.join(self.print_property(p) for p in self.attributes.values() if not self.hasattr(p.clean_name))

        return '{}{}'.format(header, properties)

    def print_property(self, prop):
        data = prop.tostring()
        if prop.name == SECONDARY_LOOKUP_FIELDS.get(self.name, '!None'):
            head, tail = data.split('\n', 1)
            data = '{}\n    @caching\n{}'.format(head, tail)
        return data

    def header(self):
        base_mixin = self.get_base_template()
        if base_mixin is not None:
            header = '# Base mixin for {}: {}\n'.format(self.python_name, base_mixin)
            header += "class {name}({base}, mixin.{mixin}):\n".format(name=self.python_name,
                                                                base=self.python_base_class,
                                                                mixin=base_mixin.__name__)
            header += '\n\n    # END HEADER\n'
        else:
            header = '# No base template found for {}\n'.format(self.python_name)
            header += "class {name}({base}):\n".format(name=self.python_name, base=self.python_base_class)
            header += '\n\n    # END HEADER\n'

        header += "    # Abstract: {}\n".format(self.abstract)
        header += "    # Simple: {}\n".format(self.simple)
        header += "    # Object class: {}\n".format(self.default_base_class)
        header += "    # Source schema: {}\n".format(self.source_schema)

        if self.display_identifier is not None:
            header += "    _DISPLAY_IDENTIFIER = '{}'\n".format(self.display_identifier)

        if 'fields' in self.attributes:
            header += "    _HAS_FIELDS = True\n"

        if self.parent_class is not None:
            header += "    #_PARENT_CLASS = {}\n".format(self.python_parent_class)
            header += "    _FIELD_NAME = '{}'\n".format(self.field_name)
        elif self.name in FIELD_HINTS:
            header += "    _CONTAINED_IN = '{}'\n".format(FIELD_HINTS[self.name])

        header += "    _XSI_TYPE = '{}'\n\n".format(self.name)

        header += "    @classmethod\n" \
                  "    def __register__(cls, target):\n" \
                  "        target['{}'] = cls\n\n".format(self.name)

        if self.name in SECONDARY_LOOKUP_FIELDS:
            header += self.init

        if self.display_identifier is not None:
            header += ("    @property\n"
                       "    def __display_identifier(self):\n"
                       "        return self.{}\n\n".format(self.display_identifier))

        return header

    # Abstract stuff that needs to be reimplemented by subclasses
    @abstractproperty
    def default_base_class(self):
        """
        The default base class if none is supplied
        """


class SimpleClassWriter(BaseClassWriter):
    @property
    def base_class(self):
        return "XNATSubObjectMixin"

    def create_listing(self, field_name, secondary_lookup):
        return """
        # Automatically generated PropertyListing, by {element_class_name} (SimpleClassWriter)
        # Secondary lookup: '{secondary_lookup}'
        return XNATSimpleListing(
            parent=self,
            field_name='{field_name}',
            secondary_lookup_field={secondary_lookup}
        )""".format(field_name=field_name,
                    secondary_lookup=secondary_lookup,
                    element_class_name=self.name)

    @property
    def default_base_class(self):
        return 'XNATSubObjectMixin'


class SubObjectClassWriter(BaseClassWriter):
    def create_listing(self, field_name, secondary_lookup):
        return """
        # Automatically generated PropertyListing, by {element_class_name} (SubObjectClassWriter)
        # Secondary lookup: '{secondary_lookup}'
        return XNATSubListing(parent=self,
                              field_name='{field_name}',
                              secondary_lookup_field={secondary_lookup},
                              xsi_type='{type_}')""".format(field_name=field_name,
                                                            secondary_lookup=secondary_lookup,
                                                            element_class_name=self.name,
                                                            type_=self.name)

    @property
    def xsi_type_registration(self):
        result = 'xnatpy:' + self.name

        return result

    @property
    def default_base_class(self):
        return 'XNATSubObjectMixin'


class NestedObjectClassWriter(BaseClassWriter):
    @property
    def default_base_class(self):
        return 'XNATNestedObjectMixin'

    @property
    def xsi_type_registration(self):
        return self.xsi_type

    def create_listing(self, field_name, secondary_lookup):
        if secondary_lookup is None:
            secondary_lookup = self.display_identifier

        if '/' in field_name:
            field_name = field_name.split('/')[0]

        return """
        # Automatically generated PropertyListing, by {element_class_name} (NestedObjectClassWriter)
        # Secondary lookup: '{secondary_lookup}'
        return XNATSubListing(uri=self.fulluri + '/{field_name}',
                                 parent=self,
                                 field_name='{field_name}',
                                 secondary_lookup_field={secondary_lookup},
                                 xsi_type='{type_}')""".format(field_name=field_name,
                                                               secondary_lookup=secondary_lookup,
                                                               element_class_name=self.name,
                                                               type_=self.name)


class ObjectClassWriter(BaseClassWriter):
    def __repr__(self):
        return '<ObjectClassWriter {}({})>'.format(self.name, self.base_class)

    def create_listing(self, field_name, secondary_lookup):
        if secondary_lookup is None:
            secondary_lookup = self.display_identifier

        if '/' in field_name:
            field_name = field_name.split('/')[0]

        return """
        # Automatically generated PropertyListing, by {element_class_name} (ObjectClassWriter)
        # Secondary lookup: '{secondary_lookup}'
        return XNATListing(uri=self.fulluri + '/{field_name}',
                           parent=self,
                           field_name='{field_name}',
                           secondary_lookup_field={secondary_lookup},
                           xsi_type='{type_}')""".format(field_name=field_name,
                                                         secondary_lookup=secondary_lookup,
                                                         element_class_name=self.name,
                                                         type_=self.name)

    @property
    def default_base_class(self):
        return 'XNATObjectMixin'

    @property
    def init(self):
        return \
"""    def __init__(self, uri=None, xnat_session=None, id_=None, datafields=None, parent=None, {lookup}=None, **kwargs):
        super({name}, self).__init__(uri=uri, xnat_session=xnat_session, id_=id_, datafields=datafields, parent=parent, {lookup}={lookup}, **kwargs)
        if {lookup} is not None:
            self._cache['{lookup}'] = {lookup}

""".format(name=self.python_name, lookup=SECONDARY_LOOKUP_FIELDS[self.name])


class AttributeWriter(BaseWriter):
    # Give easy access to prototypes attributes
    @property
    def parser(self):
        return self.prototype.parser

    @property
    def name(self):
        return self.prototype.name

    @property
    def clean_name(self):
        return self.prototype.clean_name

    @property
    def type(self):
        return self.prototype.type

    @property
    def parent_class(self):
        return self.prototype.parent_class

    @property
    def parent_property(self):
        return self.prototype.parent_property

    @property
    def restrictions(self):
        return self.prototype.restrictions

    @property
    def element_class(self):
        return self.prototype.element_class

    @property
    def min_occurs(self):
        return self.prototype.min_occur

    @property
    def max_occurs(self):
        return self.prototype.max_occur

    @property
    def docstring(self):
        return self.prototype.docstring

    @property
    def display_identifier(self):
        return self.prototype.display_identifier

    @property
    def field_name(self):
        return self.prototype.field_name

    # Actual funcions
    def __repr__(self):
        parent = self.parent_class.name if self.parent_class else None
        element = self.element_class.name if self.element_class else None

        return '<{} {} (parent: {}, element: {})>'.format(type(self).__name__,
                                                          self.name,
                                                          parent,
                                                          element)

    def restrictions_code(self):
        if len(self.restrictions) > 0:
            data = '\n        # Restrictions for value'
            if 'min' in self.restrictions:
                data += "\n        if value < {min}:\n            raise ValueError('{name} has to be greater than or equal to {min}')\n".format(name=self.name, min=self.restrictions['min'])
            if 'max' in self.restrictions:
                data += "\n        if value > {max}:\n            raise ValueError('{name} has to be smaller than or equal to {max}')\n".format(name=self.name, max=self.restrictions['max'])
            if 'maxlength' in self.restrictions:
                data += "\n        if len(value) > {maxlength}:\n            raise ValueError('length {name} has to be smaller than or equal to {maxlength}')\n".format(name=self.name, maxlength=self.restrictions['maxlength'])
            if 'minlength' in self.restrictions:
                data += "\n        if len(value) < {minlength}:\n            raise ValueError('length {name} has to be larger than or equal to {minlength}')\n".format(name=self.name, minlength=self.restrictions['minlength'])
            if 'enum' in self.restrictions:
                data += "\n        if value not in [{enum}]:\n            raise ValueError('{name} has to be one of: {enum}')\n".format(name=self.name, enum=', '.join('"{}"'.format(x.replace("'", "\\'")) for x in self.restrictions['enum']))

            return data
        else:
            return ''


class ConstantWriter(AttributeWriter):
    def __init__(self, parser, name, value=None):
        super(ConstantWriter, self).__init__(parser, name)

        self.value = value

    def __str__(self):
        return "    {s.clean_name} = {s.value}".format(s=self)

    def __repr__(self):
        return '<ConstantWriter {}({})>'.format(self.name, self.value)

    @property
    def clean_name(self):
        name = re.sub('[^0-9a-zA-Z]+', '_', self.name)
        name = '_{}'.format(name.upper())

        return name

    def tostring(self):
        return "CONSTANT = 'VALUE'"


class PropertyWriter(AttributeWriter):
    def __repr__(self):
        return '<PropertyWriter {}({})>'.format(self.name, self.type_)

    def tostring(self):
        type_string = ':py:class:`{}`'.format(full_class_name(TYPE_TO_PYTHON.get(self.type, str)))

        docstring_content = self.docstring + '\n\n' if self.docstring else ''
        docstring_content += "\n        Property of type: {}".format(type_string)
        docstring = '\n        """{}"""'.format(docstring_content)
        return \
    """    @mixedproperty
    def {clean_name}(cls):{docstring}
        # 0 Automatically generated Property, type: {type_}
        return search.SearchField(cls, "{name}")

    @{clean_name}.getter
    def {clean_name}(self):{docstring}
        # Generate automatically, type: {type_}
        return self.get("{name}", type_="{type_}")

    @{clean_name}.setter
    def {clean_name}(self, value):{docstring}{restrictions}
        # Automatically generated Property, type: {type_}
        self.set("{name}", value, type_="{type_}")
        
    @{clean_name}.deleter
    def {clean_name}(self):{docstring}
        # Automatically generated Property, type: {type_}
        self.del_("{name}")""".format(clean_name=self.clean_name,
                                      docstring=docstring,
                                      name=self.name,
                                      type_=self.type,
                                      restrictions=self.restrictions_code())


class SubObjectPropertyWriter(AttributeWriter):
    def __repr__(self):
        return '<SubObjectPropertyWriter {}({})>'.format(self.name, self.type)

    def tostring(self):
        if self.type is None:
            if self.element_class is not None:
                xsi_type = self.element_class.name
                xsi_type_arg = ', "{}"'.format(xsi_type)
            else:
                xsi_type = '_UNKNOWN_'
                xsi_type_arg = ', "{}"'.format(xsi_type)
        else:
            xsi_type = '{}'.format(core.TYPE_HINTS.get(self.name, self.type))
            xsi_type_arg = ', "{}"'.format(xsi_type)

        type_def = self.type or self.element_class

        if type_def:
            if isinstance(type_def, ClassPrototype):
                type_string = 'xnat.classes.{}'.format(pythonize_class_name(type_def.name))
            elif type_def.startswith('xs:'):
                type_string = full_class_name(TYPE_TO_PYTHON.get(type_def, str))
            else:
                type_string = type_def
                if ':' in type_string:
                    type_string.split(':', 1)
                type_string = 'xnat.classes.{}'.format(pythonize_class_name(type_string))

            type_string = ':py:class:`listing <xnat.core.XNATBaseListing>` of :py:class:`{}`'.format(type_string)
        else:
            type_string = 'Unknown'

        docstring_content = self.docstring + '\n\n' if self.docstring else '\n'
        docstring_content += "        Property of type: {}".format(type_string)
        docstring = '\n        """{}"""'.format(docstring_content)

        return \
            """    @mixedproperty
    def {clean_name}(cls):{docstring}
        # 1 Automatically generated Property, type: {type_}
        return XNAT_CLASS_LOOKUP["{xsi_type}"]

    @{clean_name}.getter
    @caching
    def {clean_name}(self):
        # Generated automatically, type: {type_}
        return self.get_object("{name}"{xsi_type_arg})""".format(clean_name=self.clean_name,
                                                                 docstring=docstring,
                                                                 name=self.name,
                                                                 type_=self.type,
                                                                 xsi_type=xsi_type,
                                                                 xsi_type_arg=xsi_type_arg)


class ListingPropertyWriter(AttributeWriter):
    def __repr__(self):
        return '<ListingPropertyWriter {}({})>'.format(self.name, self.type)

    def tostring(self):
        if self.display_identifier is not None:
            secondary_lookup = self.display_identifier
        elif self.type is not None:
            cls = self.parser.class_list.get(self.type)
            if cls is not None:
                if cls.name in SECONDARY_LOOKUP_FIELDS:
                    secondary_lookup = SECONDARY_LOOKUP_FIELDS[cls.name]
                else:
                    secondary_lookup = cls.display_identifier
            else:
                secondary_lookup = None
        else:
            secondary_lookup = None

        if secondary_lookup is not None:
            secondary_lookup = "'{}'".format(secondary_lookup)

        field_name = self.field_name or self.name

        # Get the correct type for in the docstring
        type_def = self.element_class or self.type

        if isinstance(type_def, ClassPrototype):
            type_string = 'xnat.classes.{}'.format(pythonize_class_name(type_def.name))
        elif type_def.startswith('xs:'):
            type_string = full_class_name(TYPE_TO_PYTHON.get(type_def, str))
        else:
            type_string = type_def
            if ':' in type_string:
                type_string.split(':', 1)
            type_string = 'xnat.classes.{}'.format(pythonize_class_name(type_string))

        type_string = ':py:class:`listing <xnat.core.XNATBaseListing>` of :py:class:`{}`'.format(type_string)

        docstring_content = self.docstring + '\n\n        ' if self.docstring else ''
        docstring_content += type_string
        docstring = '\n        """ {} """'.format(docstring_content)

        property_base = '''    @property
    @caching
    def {clean_name}(self):{docstring}'''.format(clean_name=self.clean_name, docstring=docstring)
        if self.element_class is not None:
            element_class = self.element_class
            property_base += element_class.writer.create_listing(secondary_lookup=secondary_lookup, field_name=field_name)
        elif not self.type.startswith('xs:'):
            element_class = self.parser.class_list[self.type]
            property_base += element_class.writer.create_listing(secondary_lookup=secondary_lookup, field_name=field_name)
        else:
            property_base += "\n        # TODO: Implement simple type listing! (type: {})\n".format(self.type) + \
                             "        pass"
        return property_base


class SchemaParser(object):
    def __init__(self, logger, debug=False):
        # Manage XML namespaces
        self.namespaces = {}
        self.namespace_prefixes = {}
        self.target_namespace = ''
        self.current_schema = None

        self.class_list = collections.OrderedDict()
        self.class_list = collections.OrderedDict()
        self.unknown_tags = set()
        self.new_class_stack = [None]
        self.new_property_stack = [None]
        self.property_prefixes = []
        self.debug = debug
        self.schemas = []
        self.class_names = collections.OrderedDict()
        self.logger = logger

    def parse_schema_xmlstring(self, xml, schema_uri):
        self.current_schema = schema_uri
        root = ElementTree.fromstring(xml)

        # Get the namespaces from the XML document (ElementTree loses them, so
        # get them by hand) and register them
        xml_schema_element_match = re.search('<xs:schema([^>]*)>', xml)

        if xml_schema_element_match is None:
            result = []
        else:
            xml_schema_element = xml_schema_element_match.group(1)
            result = re.findall(r'xmlns:(?P<prefix>\w+)="(?P<ns>\S+)"', xml_schema_element)

        for prefix, namespace in result:
            self.logger.debug('Registering namespace: {}  ->  {}'.format(
                prefix, namespace)
            )
            self.namespace_prefixes[namespace] = prefix
            self.namespaces[prefix] = namespace

        # Register schema as being loaded
        self.schemas.append(schema_uri)

        # Parse xml schema
        self.parse(root, toplevel=True)

        if self.debug:
            self.logger.debug('Found {} unknown tags: {}'.format(len(self.unknown_tags),
                                                                 self.unknown_tags))

        self.current_schema = None
        return True

    def parse_schema_file(self, filepath):
        filepath = os.path.abspath(filepath)
        filepath = os.path.normpath(filepath)

        schema_uri = 'file://{}'.format(filepath)

        with open(filepath) as fin:
            data = fin.read()

        self.parse_schema_xmlstring(data, schema_uri=schema_uri)

    def parse_schema_uri(self, xnat_session, schema_uri):
        self.logger.info('=== Retrieving schema from {} ==='.format(schema_uri))

        resp = xnat_session.get(schema_uri, headers={'Accept-Encoding': None})
        data = resp.text

        try:
            return self.parse_schema_xmlstring(data, schema_uri=schema_uri)
        except ElementTree.ParseError as exception:
            if 'action="/j_spring_security_check"' in data:
                self.logger.error('You do not have access to this XNAT server, please check your credentials!')
            elif 'java.lang.IllegalStateException' in data:
                self.logger.error('The server returned an error. You probably do not'
                                  ' have access to this XNAT server, please check your credentials!')
            else:
                self.logger.info('Could not parse schema from {}, no valid XML found'.format(schema_uri))

                if self.debug:
                    self.logger.debug('XML schema request returned the following response: [{}] {}'.format(resp.status_code,
                                                                                                           data))
            return False

    @staticmethod
    def find_schema_uris(text):
        try:
            root = ElementTree.fromstring(text)
        except ElementTree.ParseError:
            raise ValueError('Could not parse xml file')

        schemas_string = root.attrib.get('{http://www.w3.org/2001/XMLSchema-instance}schemaLocation', '')
        schemas = [x for x in schemas_string.split() if x.endswith('.xsd')]

        return schemas

    def __iter__(self):
        visited = set()
        nr_previsited = len(visited)
        tries = 0
        yielded_anything = True
        while len(visited) < len(self.class_list) and yielded_anything and tries < 25:
            yielded_anything = False
            for key, value in self.class_list.items():
                if key in visited:
                    continue

                base = value.base_class
                if base is not None and not base.startswith('xs:') and base not in visited:
                    if self.debug:
                        self.logger.debug("Wait with processing {} because base {} is not yet processed".format(
                            value.name, base
                        ))
                    continue

                if value.parent_class is not None and value.parent_class not in visited:
                    if self.debug:
                        self.logger.debug("Wait with processing {} because parent {} is not yet processed".format(
                            value.name, value.parent_class
                        ))
                    continue

                visited.add(key)
                yielded_anything = True
                self.logger.info('Processing {} (base class {})'.format(value.name, value.base_class))
                yield value

            tries += 1

        expected = len(self.class_list) + nr_previsited  # We started with two "visited" classes
        if self.debug:  # and len(visited) < len(self.class_list):
            missed = sorted(set(self.class_list) - visited)
            if self.debug:
                self.logger.debug('Visited: {}, expected: {}'.format(len(visited), expected))
                self.logger.debug('Visited: {}'.format(visited))
                self.logger.info('Missed: {}'.format(missed))
                self.logger.info('Missed base class: {}'.format([self.class_list[x].base_class for x in missed]))
                self.logger.debug('Spent {} iterations'.format(tries))

    @contextlib.contextmanager
    def _descend(self, new_class=None, new_property=None, property_prefix=None):
        if new_class is not None:
            self.new_class_stack.append(new_class)
        if new_property is not None:
            self.new_property_stack.append(new_property)
        if property_prefix is not None:
            self.property_prefixes.append(property_prefix)

        yield

        if new_class is not None:
            self.new_class_stack.pop()
        if new_property is not None:
            self.new_property_stack.pop()
        if property_prefix is not None:
            self.property_prefixes.pop()

    @property
    def _current_class(self):
        return self.new_class_stack[-1]

    @property
    def _current_property(self):
        return self.new_property_stack[-1]

    def parse(self, element, toplevel=False):
        if element.tag in self.PARSERS:
            self.PARSERS[element.tag](self, element)
        else:
            self._parse_unknown(element)

    # TODO: We should check the following restrictions:
    # http://www.w3schools.com/xml/schema_facets.asp

    def _parse_all(self, element):
        self._parse_children(element)

    def _parse_annotation(self, element):
        self._parse_children(element)

    def _parse_attribute(self, element):
        name = element.get('name')
        type_ = element.get('type')

        if self._current_class is not None:
            if name is None:
                if self.debug:
                    self.logger.warning('Encountered attribute without name')
                return

            new_property = AttributePrototype(self, name=name, logger=self.logger,
                                              type=type_, parent_class=self._current_class,
                                              parent_property=self._current_property)

            self._current_class.attributes[name] = new_property

            with self._descend(new_property=new_property):
                self._parse_children(element)

    def _parse_children(self, element):
        for child in list(element):
            self.parse(child)

    def _parse_choice(self, element):
        self._parse_children(element)

    def _parse_complex_content(self, element):
        self._parse_children(element)

    def _parse_complex_type(self, element):
        name = element.get('name')
        parent_class = None
        field_name = None

        if name is None:
            name = 'xnatpy:' + self._current_class.name.split(":", 1)[-1] + self._current_property.name.capitalize()
            parent_class = self._current_class.name
            field_name = self._current_property.name
        else:
            if ':' not in name:
                name = self.target_namespace_prefix + name

        new_class = ClassPrototype(self,
                                   name=name,
                                   logger=self.logger,
                                   field_name=field_name,
                                   parent_class=parent_class)

        if self._current_property is not None:
            self._current_property.element_class = new_class
            new_class.parent_property = self._current_property

        self.class_list[name] = new_class

        # Descend
        with self._descend(new_class=new_class):
            self._parse_children(element)

    def _parse_documentation(self, element):
        if self._current_property is not None:
            self._current_property.docstring = element.text

    def _parse_element(self, element):
        name = element.get('name')
        type_ = element.get('type')

        if name is None:
            abstract = element.get('abstract')
            if abstract is not None:
                self._current_class.abstract = abstract == "true"
            else:
                if self.debug:
                    self.logger.warning('Encountered attribute without name')
            return

        if self._current_class is not None:
            min_occur = element.get('minOccurs')
            max_occur = element.get('maxOccurs')

            new_property = AttributePrototype(self, name=name, logger=self.logger,
                                              type=type_, parent_class=self._current_class,
                                              parent_property=self._current_property,
                                              min_occur=min_occur, max_occur=max_occur)

            self._current_class.attributes[name] = new_property

            with self._descend(new_property=new_property):
                self._parse_children(element)

    def _parse_enumeration(self, element):
        if 'enum' in self._current_property.restrictions:
            self._current_property.restrictions['enum'].append(element.get('value'))
        else:
            self._current_property.restrictions['enum'] = [element.get('value')]

    def _parse_error(self, element):
        raise NotImplementedError('The parser for {} has not yet been implemented'.format(element.tag))

    def _parse_extension(self, element):
        new_base = element.get('base')
        if new_base.startswith('xs:'):
            # Need to create a base object as we do not have this
            previous_base = new_base
            name = new_base[3:].capitalize()
            if self._current_property is not None:
                new_prop = self._current_property.name
                new_base = 'xnatpy:{}{}'.format(new_prop, name)
            else:
                new_prop = self._current_class.name
                new_base = self._current_class.name + name
                new_base = 'xnatpy:{}'.format(new_base.split(':', 1)[1])

            new_base_class = ClassPrototype(self,
                                            name=new_base,
                                            logger=self.logger,
                                            simple=True)

            new_base_class.attributes[new_prop] = AttributePrototype(self,
                                                                     name=new_prop,
                                                                     logger=self.logger,
                                                                     type=previous_base)
            self.class_list[new_base] = new_base_class

        self._current_class.base_class = new_base
        self._parse_children(element)

    def _parse_ignore(self, element):
        pass

    def _parse_max_inclusive(self, element):
        self._current_property.restrictions['max'] = element.get('value')

    def _parse_max_length(self, element):
        self._current_property.restrictions['maxlength'] = element.get('value')

    def _parse_min_inclusive(self, element):
        self._current_property.restrictions['min'] = element.get('value')

    def _parse_min_length(self, element):
        self._current_property.restrictions['minlength'] = element.get('value')

    def _parse_restriction(self, element):
        old_type = self._current_property.type
        new_type = element.get('base')

        if old_type is not None and old_type != new_type:
            raise ValueError('Trying to override a type from a restriction!? (from {} to {})'.format(old_type, new_type))

        self._current_property.type = new_type

        # Parse further restrictions
        self._parse_children(element)

    def _parse_schema(self, element):
        self.target_namespace = element.get("targetNamespace", '')
        if self.target_namespace != '':
            self.target_namespace_prefix = self.namespace_prefixes[self.target_namespace] + ':'
        else:
            self.target_namespace_prefix = ''

        for child in list(element):
            if child.tag in [
                '{http://www.w3.org/2001/XMLSchema}complexType',
                '{http://www.w3.org/2001/XMLSchema}simpleType'
            ]:
                self.parse(child)
            elif child.tag == '{http://www.w3.org/2001/XMLSchema}element':
                name = child.get('name')
                type_ = child.get('type')

                if self.debug:
                    self.logger.debug('Adding {} -> {} to class name map'.format(name, type_))
                self.class_names[type_] = name
            else:
                if self.debug:
                    self.logger.debug('Skipping non-class top-level tag {}'.format(child.tag))

    def _parse_sequence(self, element):
        self._parse_children(element)

    def _parse_simple_content(self, element):
        self._parse_children(element)

    def _parse_simple_type(self, element):
        name = element.get("name")

        # This is not top-level in the schema and just adds restriction
        if name is None:
            self._parse_children(element)
            return

        # This is the top-level of schema and a sort of typedef
        if ':' not in name:
            name = self.target_namespace_prefix + name

        new_class = ClassPrototype(self,
                                   name=name,
                                   logger=self.logger,
                                   simple=False)

        new_property = AttributePrototype(self,
                                          name="value",
                                          logger=self.logger,
                                          type=None)
        new_class.attributes["value"] = new_property

        self.class_list[name] = new_class

        # Descend
        with self._descend(new_class=new_class, new_property=new_property):
            self._parse_children(element)

    def _parse_unknown(self, element):
        self.unknown_tags.add(element.tag)

    def _parse_xdat_element(self, element):
        abstract = element.get("abstract")
        if abstract is not None:
            self._current_class.abstract = abstract == "true"

        display_identifier = element.get("displayIdentifiers")
        if display_identifier is not None:
            if self._current_property is None:
                self._current_class.display_identifier = display_identifier
            else:
                self._current_property.display_identifier = display_identifier

    def _parse_sqlfield(self, element):
        if self.debug:
            self.logger.debug("CLASS: {}, PROP: {}, ELEMENT: {}".format(self._current_class.name, self._current_property.name, element))

    PARSERS = {
        '{http://www.w3.org/2001/XMLSchema}all': _parse_all,
        '{http://www.w3.org/2001/XMLSchema}annotation': _parse_annotation,
        '{http://www.w3.org/2001/XMLSchema}appinfo': _parse_children,
        '{http://www.w3.org/2001/XMLSchema}attribute': _parse_attribute,
        '{http://www.w3.org/2001/XMLSchema}attributeGroup': _parse_error,
        '{http://www.w3.org/2001/XMLSchema}choice': _parse_choice,
        '{http://www.w3.org/2001/XMLSchema}complexContent': _parse_complex_content,
        '{http://www.w3.org/2001/XMLSchema}complexType': _parse_complex_type,
        '{http://www.w3.org/2001/XMLSchema}documentation': _parse_documentation,
        '{http://www.w3.org/2001/XMLSchema}element': _parse_element,
        '{http://www.w3.org/2001/XMLSchema}enumeration': _parse_enumeration,
        '{http://www.w3.org/2001/XMLSchema}extension': _parse_extension,
        '{http://www.w3.org/2001/XMLSchema}import': _parse_ignore,
        '{http://www.w3.org/2001/XMLSchema}group': _parse_error,
        '{http://www.w3.org/2001/XMLSchema}maxInclusive': _parse_max_inclusive,
        '{http://www.w3.org/2001/XMLSchema}maxLength': _parse_max_length,
        '{http://www.w3.org/2001/XMLSchema}minInclusive': _parse_min_inclusive,
        '{http://www.w3.org/2001/XMLSchema}minLength': _parse_min_length,
        '{http://www.w3.org/2001/XMLSchema}restriction': _parse_restriction,
        '{http://www.w3.org/2001/XMLSchema}schema': _parse_schema,
        '{http://www.w3.org/2001/XMLSchema}sequence': _parse_sequence,
        '{http://www.w3.org/2001/XMLSchema}simpleContent': _parse_simple_content,
        '{http://www.w3.org/2001/XMLSchema}simpleType': _parse_simple_type,
        '{http://www.w3.org/2001/XMLSchema}unique': _parse_ignore,
        '{http://nrg.wustl.edu/xdat}element': _parse_xdat_element,
        '{http://nrg.wustl.edu/xdat}field': _parse_children,
        '{http://nrg.wustl.edu/xdat}sqlField': _parse_sqlfield,
    }

    def prune_tree(self):
        to_remove = set()

        for cls in self.class_list.values():
            for property_key, prop in cls.attributes.items():
                element_class = prop.element_class

                if element_class is not None and len(element_class.attributes) == 1:
                    element_property = next(iter(element_class.attributes.values()))

                    if element_property.property_type == 'listing':
                        # Reset parent to current parent
                        parent_class = prop.parent_class

                        element_property.parent_class = parent_class
                        element_property.field_name = '{}/{}'.format(prop.name, element_property.name)
                        element_property.name = prop.name

                        # The new element_class has to be updated to take the place of the old element_class
                        new_element_class = element_property.element_class
                        if new_element_class is not None:
                            new_element_class.name = element_class.name
                            new_element_class.field_name = '{}/{}'.format(element_class.field_name,
                                                                          new_element_class.field_name)
                            new_element_class.parent_class = element_class.parent_class

                        if element_class.name in self.class_list:
                            if self.debug:
                                self.logger.info('Removing class {} from parser, it only has one element!'.format(
                                    element_class.name)
                                )
                            to_remove.add(element_class.name)

                        cls.attributes[property_key] = element_property
                    elif self.debug:
                        self.logger.debug("Ignoring non-listing...")
                        self.logger.debug("Element class: {}".format(element_class.__dict__))

        for key in to_remove:
            del self.class_list[key]

        for cls in self.class_list.values():
            for property_key, prop in cls.attributes.items():
                # Only consider simplifying listings
                if prop.property_type != 'listing':
                    continue

                # If the type is not by element class, no need to check further
                if prop.element_class is None:
                    continue

                if prop.element_class.simple:
                    if self.debug:
                        self.logger.debug('Found simple mapping {}.{} -> {}'.format(cls.name,
                                                                                        property_key,
                                                                                        prop.element_class.name))

                    if len(prop.element_class.attributes) > 2:
                        if self.debug:
                            self.logger.debug(
                                'Too many attributes to simplify (found {} attributed)'.format(
                                    len(prop.element_class.attributes)
                                )
                            )
                        continue

                # Attempt to find listings with simple type

    def write(self, code_file):
        if self.debug:
            self.logger.debug('namespaces: {}'.format(self.namespaces))
            self.logger.debug('namespace prefixes: {}'.format(self.namespace_prefixes))

        self.logger.info('=== Pruning data structure ===')
        before = set(self.class_list.keys())
        if self.debug:
            self.logger.debug('Classed before prune: {}'.format(before))
        self.prune_tree()
        after = set(self.class_list.keys())
        if self.debug:
            self.logger.debug('Classed after prune: {}'.format(after))
            self.logger.info('Classes removed by pruning: {}'.format(sorted(before - after)))

        self.logger.info('=== Writing result ===')
        schemas = '\n'.join('# - {}'.format(s) for s in self.schemas)
        code_file.write(FILE_HEADER.format(schemas=schemas,
                                           file_secondary_lookup=SECONDARY_LOOKUP_FIELDS['xnat:fileData']))

        code_file.write('\n\n\n'.join(c.tostring().strip() for c in self if c.name is not None))



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import unicode_literals
import binascii
import os
import sys
from setuptools import setup

# Get information about the version (polling mercurial if possible)
version = '0.3.23'
dirstate = './.hg/dirstate'

if __name__ == '__main__':
    # Get the requirements
    with open('requirements.txt', 'r') as fh:
        _requires = fh.read().splitlines()

    if os.path.isfile(dirstate):
        with open(dirstate, 'rb') as f_dirstate:
            hg_version = binascii.hexlify(f_dirstate.read(20))

            if sys.version_info[0] >= 3:
                hg_version = hg_version.decode('utf-8')
    else:
        hg_version = None

    branch = './.hg/branch'
    if os.path.isfile(branch):
        with open(branch, 'r') as f_branch:
            hg_branch = f_branch.read().strip()
    else:
        hg_branch = ''

    if hg_version is not None:
        extra_version = '{}-{}-{}'.format(version, hg_branch, hg_version[0:6])
    else:
        extra_version = '{}'.format(version)

    # Write information to version.py
    with open('./xnat/version.py', 'w') as f_version:
        f_version.write('version = "{}"\n'.format(version))
        f_version.write('extra_version = "{}"\n'.format(extra_version))
        f_version.write('hg_revision = "{}"\n'.format(hg_version))
        f_version.write('hg_branch = "{}"\n'.format(hg_branch))

    # Set the entry point
    entry_points = {
        "console_scripts": [
            "xnat_cp_project = xnat.scripts.copy_project:main",
        ]
    }

    setup(
        name='xnat',
        version=version,
        author='H.C. Achterberg',
        author_email='hakim.achterberg@gmail.com',
        packages=[str('xnat'), str('xnat.scripts')],
        url='https://gitlab.com/radiology/infrastructure/xnatpy',
        license='Apache 2.0',
        description='An XNAT client that exposes the XNAT REST interface as python objects. Part of the interface is automatically generated based on the servers data model as defined by the xnat schema.',
        long_description=open('README.rst').read(),
        install_requires=_requires,
        entry_points=entry_points,
        classifiers=[
            "Development Status :: 4 - Beta",
            "Intended Audience :: Developers",
            "Intended Audience :: Healthcare Industry",
            "Intended Audience :: Science/Research",
            "License :: OSI Approved :: Apache Software License",
            "Natural Language :: English",
            "Programming Language :: Python :: 2.7",
            "Programming Language :: Python :: 3.4",
            "Programming Language :: Python :: 3.5",
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: 3.7",
            "Topic :: Scientific/Engineering :: Bio-Informatics",
            "Topic :: Scientific/Engineering :: Medical Science Apps.",
            ]

    )



# Copyright 2011-2015 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import unicode_literals
import datetime
import re

import isodate

from .core import XNATBaseObject, caching
from .datatypes import to_date, to_time
from .utils import RequestsFileLike

try:
    PYDICOM_LOADED = True
    import pydicom
except ImportError:
    PYDICOM_LOADED = False


class PrearchiveSession(XNATBaseObject):
    @property
    def id(self):
        """
        A unique ID for the session in the prearchive
        :return:
        """
        return '{}/{}/{}'.format(self.data['project'], self.data['timestamp'], self.data['name'])

    @property
    def xpath(self):
        return "xnatpy:prearchiveSession"

    @property
    def fulldata(self):
        # There is a bug in 1.7.0-1.7.2 that misses a route in the REST API
        # this should be fixed from 1.7.3 onward
        if re.match('^1\.7\.[0-2]', self.xnat_session.xnat_version):
            # Find the xnat prearchive project uri
            project_uri = self.uri.rsplit('/', 2)[0]

            # We need to search for session with url field without the /data start
            target_uri = self.uri[5:] if self.uri.startswith('/data') else self.uri
            all_sessions = self.xnat_session.get_json(project_uri)
            for session in all_sessions['ResultSet']['Result']:
                if session['url'] == target_uri:
                    return session
            else:
                raise IndexError('Could not find specified prearchive session {}'.format(self.uri))
        else:
            return self.xnat_session.get_json(self.uri)['ResultSet']['Result'][0]

    @property
    @caching
    def data(self):
        return self.fulldata

    @property
    def autoarchive(self):
        return self.data['autoarchive']

    @property
    def folder_name(self):
        return self.data['folderName']

    @property
    def lastmod(self):
        lastmod_string = self.data['lastmod']
        return datetime.datetime.strptime(lastmod_string, '%Y-%m-%d %H:%M:%S.%f')

    @property
    def name(self):
        return self.data['name']

    @property
    def label(self):
        return self.name

    @property
    def prevent_anon(self):
        return self.data['prevent_anon']

    @property
    def prevent_auto_commit(self):
        return self.data['prevent_auto_commit']

    @property
    def project(self):
        return self.data['project']

    @property
    def scan_date(self):
        try:
            return to_date(self.data['scan_date'])
        except isodate.ISO8601Error:
            return None

    @property
    def scan_time(self):
        try:
            return to_time(self.data['scan_time'])
        except isodate.ISO8601Error:
            return None

    @property
    def status(self):
        return self.data['status']

    @property
    def subject(self):
        return self.data['subject']

    @property
    def tag(self):
        return self.data['tag']

    @property
    def timestamp(self):
        return self.data['timestamp']

    @property
    def uploaded(self):
        """
        Datetime when the session was uploaded
        """
        uploaded_string = self.data['uploaded']
        try:
            return datetime.datetime.strptime(uploaded_string, '%Y-%m-%d %H:%M:%S.%f')
        except ValueError:
            return None

    @property
    def scans(self):
        """
        List of scans in the prearchive session
        """
        data = self.xnat_session.get_json(self.uri + '/scans')
        # We need to prepend /data to our url (seems to be a bug?)

        return [PrearchiveScan('{}/scans/{}'.format(self.uri, x['ID']),
                               self.xnat_session,
                               id_=x['ID'],
                               datafields=x) for x in data['ResultSet']['Result']]

    def download(self, path):
        """
        Method to download the zip of the prearchive session

        :param str path: path to download to
        :return: path of the downloaded zip file
        :rtype: str
        """
        self.xnat_session.download_zip(self.uri, path)
        return path

    def archive(self, overwrite=None, quarantine=None, trigger_pipelines=None,
                project=None, subject=None, experiment=None):
        """
        Method to archive this prearchive session to the main archive

        :param str overwrite: how the handle existing data (none, append, delete)
        :param bool quarantine: flag to indicate session should be quarantined
        :param bool trigger_pipelines: indicate that archiving should trigger pipelines
        :param str project: the project in the archive to assign the session to
        :param str subject: the subject in the archive to assign the session to
        :param str experiment: the experiment in the archive to assign the session content to
        :return: the newly created experiment
        :rtype: xnat.classes.ExperimentData
        """
        query = {'src': self.uri}

        if overwrite is not None:
            if overwrite not in ['none', 'append', 'delete']:
                raise ValueError('Overwrite should be none, append or delete!')
            query['overwrite'] = overwrite

        if quarantine is not None:
            if isinstance(quarantine, bool):
                if quarantine:
                    query['quarantine'] = 'true'
                else:
                    query['quarantine'] = 'false'
            else:
                raise TypeError('Quarantine should be a boolean')

        if trigger_pipelines is not None:
            if isinstance(trigger_pipelines, bool):
                if trigger_pipelines:
                    query['triggerPipelines'] = 'true'
                else:
                    query['triggerPipelines'] = 'false'
            else:
                raise TypeError('trigger_pipelines should be a boolean')

        # Change the destination of the session
        # BEWARE the dest argument is completely ignored, but there is a work around:
        # HACK: See https://groups.google.com/forum/#!searchin/xnat_discussion/prearchive$20archive$20service/xnat_discussion/hwx3NOdfzCk/rQ6r2lRpZjwJ
        if project is not None:
            query['project'] = project

        if subject is not None:
            query['subject'] = subject

        if experiment is not None:
            query['session'] = experiment

        response = self.xnat_session.post('/data/services/archive', query=query)
        object_uri = response.text.strip()

        self.clearcache()  # Make object unavailable
        return self.xnat_session.create_object(object_uri)

    def delete(self, asynchronous=None):
        """
        Delete the session from the prearchive

        :param bool asynchronous: flag to delete asynchronously
        :return: requests response
        """
        query = {'src': self.uri}

        if asynchronous is not None:
            if isinstance(asynchronous, bool):
                if asynchronous:
                    query['async'] = 'true'
                else:
                    query['async'] = 'false'
            else:
                raise TypeError('async should be a boolean')

        response = self.xnat_session.post('/data/services/prearchive/delete', query=query)
        self.clearcache()
        return response

    def rebuild(self, asynchronous=None):
        """
        Rebuilt the session in the prearchive

        :param bool asynchronous: flag to rebuild asynchronously
        :return: requests response
        """
        query = {'src': self.uri}

        if asynchronous is not None:
            if isinstance(asynchronous, bool):
                if asynchronous:
                    query['async'] = 'true'
                else:
                    query['async'] = 'false'
            else:
                raise TypeError('async should be a boolean')

        response = self.xnat_session.post('/data/services/prearchive/rebuild', query=query)
        self.clearcache()
        return response

    def move(self, new_project, asynchronous=None):
        """
        Move the session to a different project in the prearchive

        :param str new_project: the id of the project to move to
        :param bool asynchronous: flag to move asynchronously
        :return: requests response
        """
        query = {'src': self.uri,
                 'newProject': new_project}

        if asynchronous is not None:
            if isinstance(asynchronous, bool):
                if asynchronous:
                    query['async'] = 'true'
                else:
                    query['async'] = 'false'
            else:
                raise TypeError('async should be a boolean')

        response = self.xnat_session.post('/data/services/prearchive/move', query=query)
        self.clearcache()
        return response


class PrearchiveScan(XNATBaseObject):
    def __init__(self, uri, xnat_session, id_=None, datafields=None, parent=None, fieldname=None):
        super(PrearchiveScan, self).__init__(uri=uri,
                                             xnat_session=xnat_session,
                                             id_=id_,
                                             datafields=datafields,
                                             parent=parent,
                                             fieldname=fieldname)

        self._fulldata = {'data_fields': datafields}

    @property
    def series_description(self):
        """
        The series description of the scan
        """
        return self.data['series_description']

    @property
    def files(self):
        """
        List of files contained in the scan
        """
        data = self.xnat_session.get_json(self.uri + '/resources/DICOM/files')

        return [PrearchiveFile(x['URI'],
                               self.xnat_session,
                               id_=x['Name'],
                               datafields=x) for x in data['ResultSet']['Result']]

    def download(self, path):
        """
        Download the scan as a zip

        :param str path: the path to download to
        :return: the path of the downloaded file
        :rtype: str
        """
        self.xnat_session.download_zip(self.uri, path)
        return path

    @property
    def data(self):
        return self.fulldata['data_fields']

    @property
    def fulldata(self):
        return self._fulldata

    @property
    def xpath(self):
        return "xnatpy:prearchiveScan"

    def dicom_dump(self, fields=None):
        """
        Retrieve a dicom dump as a JSON data structure
        See the XAPI documentation for more detailed information: `DICOM Dump Service <https://wiki.xnat.org/display/XAPI/DICOM+Dump+Service+API>`_

        :param list fields: Fields to filter for DICOM tags. It can either a tag name or tag number in the format GGGGEEEE (G = Group number, E = Element number)
        :return: JSON object (dict) representation of DICOM header
        :rtype: dict
        """

        # Get the uri in the following format /prearchive/projects/${project}/${timestamp}/${session}
        # Get the uri and remove the first five characters: /data
        uri = self.uri[5:]
        return self.xnat_session.services.dicom_dump(src=uri, fields=fields)

    def read_dicom(self, file=None, read_pixel_data=False, force=False):
        # Check https://gist.github.com/obskyr/b9d4b4223e7eaf4eedcd9defabb34f13 for partial loading?
        if not PYDICOM_LOADED:
            raise RuntimeError('Cannot read DICOM, missing required dependency: pydicom')

        if file is None:
            dicom_files = sorted(self.files, key=lambda x: x.name)
            file = dicom_files[0]
        else:
            if file not in self.files:
                raise ValueError('File {} not part of scan {} DICOM resource'.format(file, self))

        with file.open() as dicom_fh:
            dicom_data = pydicom.dcmread(dicom_fh,
                                         stop_before_pixels=not read_pixel_data,
                                         force=force)

        return dicom_data


class PrearchiveFile(XNATBaseObject):
    def __init__(self, uri, xnat_session, id_=None, datafields=None, parent=None, fieldname=None):
        super(PrearchiveFile, self).__init__(uri=uri,
                                             xnat_session=xnat_session,
                                             id_=id_,
                                             datafields=datafields,
                                             parent=parent,
                                             fieldname=fieldname)

        self._fulldata = datafields

    def open(self):
        uri = self.xnat_session.url_for(self)
        request = self.xnat_session.interface.get(uri, stream=True)
        return RequestsFileLike(request)

    @property
    def data(self):
        return self.fulldata

    @property
    def fulldata(self):
        return self._fulldata

    @property
    def name(self):
        """
        The name of the file
        """
        return self.data['Name']

    @property
    def size(self):
        """
        The size of the file
        """
        return self.data['Size']

    @property
    def xpath(self):
        return "xnatpy:prearchiveFile"

    def download(self, path):
        """
        Download the file

        :param str path: the path to download to
        :return: the path of the downloaded file
        :rtype: str
        """
        self.xnat_session.download_zip(self.uri, path)
        return path


class Prearchive(object):
    def __init__(self, xnat_session):
        self._xnat_session = xnat_session
        self._cache = {}
        self._caching = True

    @property
    def caching(self):
        if self._caching is not None:
            return self._caching
        else:
            return self.xnat_session.caching

    @caching.setter
    def caching(self, value):
        self._caching = value

    @caching.deleter
    def caching(self):
        self._caching = None

    @property
    def xnat_session(self):
        return self._xnat_session

    def sessions(self, project=None):
        """
        Get the session in the prearchive, optionally filtered by project. This
        function is not cached and returns the results of a query at each call.

        :param str project: the project to filter on
        :return: list of prearchive session found
        :rtype: list
        """
        if project is None:
            uri = '/data/prearchive/projects'
        else:
            uri = '/data/prearchive/projects/{}'.format(project)

        data = self.xnat_session.get_json(uri)
        # We need to prepend /data to our url (seems to be a bug?)

        result = []
        for session_data in data['ResultSet']['Result']:
            uri = '/data{}'.format(session_data['url'])

            session = self._cache.get(uri, None)

            if session is None:
                session = PrearchiveSession(uri, self.xnat_session)
                self._cache[uri] = session

            result.append(session)

        return result

    def find(self, project=None, subject=None, session=None, status=None):
        """
        Find specific session(s) given the project/subject/session/status

        :param str project:
        :param str subject:
        :param str session:
        :param str status:
        :return: list of matching sessions
        :rtype: list[PrearchiveSession]
        """
        result = []
        sessions = self.sessions(project=project)

        for option in sessions:
            if subject is not None and option.subject != subject:
                continue

            if session is not None and option.label != session:
                continue

            if status is not None and option.status != status:
                continue

            result.append(option)

        return result
